"""
[StableHLO](https://github.com/openxla/stablehlo/blob/main/docs/spec.md)
is an operation set for high-level operations (HLO) in machine learning (ML) models.
StableHLO works as a portability layer between different ML frameworks and ML compilers:
ML frameworks that produce StableHLO programs are compatible with ML compilers that consume StableHLO programs.
"""

import abc
from collections.abc import Sequence
from typing import Annotated, ClassVar, TypeAlias, cast

from xdsl.dialects.builtin import (
    I32,
    I64,
    AnyFloat,
    AnyTensorType,
    AnyTensorTypeConstr,
    ArrayAttr,
    ComplexType,
    DenseArrayBase,
    IntegerAttr,
    IntegerType,
    TensorType,
    i64,
)
from xdsl.ir import (
    Attribute,
    Dialect,
    EnumAttribute,
    ParametrizedAttribute,
    Region,
    SpacedOpaqueSyntaxAttribute,
    SSAValue,
    StrEnum,
    TypeAttribute,
)
from xdsl.irdl import (
    BaseAttr,
    ConstraintVar,
    IRDLOperation,
    ParameterDef,
    VarConstraint,
    attr_def,
    base,
    irdl_attr_definition,
    irdl_op_definition,
    operand_def,
    result_def,
    traits_def,
    var_operand_def,
    var_region_def,
    var_result_def,
)
from xdsl.parser import AttrParser
from xdsl.printer import Printer
from xdsl.traits import IsTerminator
from xdsl.utils.exceptions import VerifyException

IntegerTensorType: TypeAlias = TensorType[IntegerType]
FloatOrComplexType: TypeAlias = AnyFloat | ComplexType
FloatOrComplexTensorType: TypeAlias = TensorType[FloatOrComplexType]
FloatTensorType: TypeAlias = TensorType[AnyFloat]

# TODO: Change to SI32 once StableHLO adopts signful integer semantics
# See: https://github.com/openxla/stablehlo/issues/22
# https://github.com/openxla/stablehlo/issues/2489
SI32TensorType: TypeAlias = TensorType[I32]


# region Abstract Base Classes


# TODO: Abstract the constraint
class ElementwiseBinaryOperation(IRDLOperation, abc.ABC):
    # TODO: Remove this constraint for complex types.
    T: ClassVar = VarConstraint("T", base(AnyTensorType))

    lhs = operand_def(T)
    rhs = operand_def(T)

    result = result_def(T)

    def __init__(
        self, lhs: SSAValue, rhs: SSAValue, result_type: Attribute | None = None
    ):
        if result_type is None:
            result_type = lhs.type
        super().__init__(operands=(lhs, rhs), result_types=(result_type,))


class IntegerTensorLikeElementwiseBinaryOperation(IRDLOperation, abc.ABC):
    T: ClassVar = VarConstraint("T", base(IntegerTensorType))

    lhs = operand_def(T)
    rhs = operand_def(T)

    result = result_def(T)

    def __init__(
        self, lhs: SSAValue, rhs: SSAValue, result_type: Attribute | None = None
    ):
        if result_type is None:
            result_type = lhs.type
        super().__init__(operands=(lhs, rhs), result_types=(result_type,))


class IntegerTensorLikeElementwiseUnaryOperation(IRDLOperation, abc.ABC):
    T: ClassVar = VarConstraint("T", base(IntegerTensorType))

    operand = operand_def(T)
    result = result_def(T)

    def __init__(self, operand: SSAValue, result_type: Attribute | None = None):
        if result_type is None:
            result_type = operand.type
        super().__init__(operands=(operand,), result_types=(result_type,))


class FloatOrComplexTensorLikeElementwiseBinaryOperation(IRDLOperation, abc.ABC):
    T: ClassVar = VarConstraint("T", base(FloatOrComplexTensorType))

    lhs = operand_def(T)
    rhs = operand_def(T)

    result = result_def(T)

    def __init__(
        self, lhs: SSAValue, rhs: SSAValue, result_type: Attribute | None = None
    ):
        if result_type is None:
            result_type = lhs.type
        super().__init__(operands=(lhs, rhs), result_types=(result_type,))


class FloatOrComplexTensorLikeElementwiseUnaryOperation(IRDLOperation, abc.ABC):
    T: ClassVar = VarConstraint("T", base(FloatOrComplexTensorType))

    operand = operand_def(T)
    result = result_def(T)

    def __init__(self, operand: SSAValue, result_type: Attribute | None = None):
        if result_type is None:
            result_type = operand.type
        super().__init__(operands=(operand,), result_types=(result_type,))


class FloatTensorLikeElementwiseUnaryOperation(IRDLOperation, abc.ABC):
    T: ClassVar = VarConstraint("T", base(FloatTensorType))

    operand = operand_def(T)
    result = result_def(T)

    def __init__(self, operand: SSAValue, result_type: Attribute | None = None):
        if result_type is None:
            result_type = operand.type
        super().__init__(operands=(operand,), result_types=(result_type,))


# endregion

# region Attributes


class ComparisonDirection(StrEnum):
    """
    Comparison direction for stablehlo.
    """

    EQ = "EQ"
    NE = "NE"
    GE = "GE"
    GT = "GT"
    LE = "LE"
    LT = "LT"


@irdl_attr_definition
class ComparisonDirectionAttr(
    EnumAttribute[ComparisonDirection], SpacedOpaqueSyntaxAttribute
):
    """
    The values of `comparison_direction` and `compare_type` have the following semantics:
    For boolean and integer element types:

    * `EQ`: lhs = rhs.
    * `NE`: lhs != rhs.
    * `GE`: lhs >= rhs.
    * `GT`: lhs > rhs.
    * `LE`: lhs <= rhs
    * `LT`: lhs < rhs.

    For floating-point element types with `compare_type = FLOAT`, the op implements the following IEEE-754 operations:

    * `EQ`: compareQuietEqual.
    * `NE`: compareQuietNotEqual.
    * `GE`: compareQuietGreaterEqual.
    * `GT`: compareQuietGreater.
    * `LE`: compareQuietLessEqual.
    * `LT`: compareQuietLess.

    For floating-point element types with `compare_type = TOTALORDER`,
    the op uses the combination of totalOrder and compareQuietEqual operations from IEEE-754.
    For complex element types,
    lexicographic comparison of (real, imag) pairs is performed using the provided `comparison_direction` and `compare_type`.
    Imposing an ordering on complex numbers involves surprising semantics,
    so in the future we are planning to remove support for complex numbers when comparison_direction is `GE`, `GT`, `LE` or `LT`.

    For quantized types, performs `dequantize_compare(lhs, rhs, comparison_direction)`

    See external [documentation](https://github.com/openxla/stablehlo/blob/main/docs/spec.md#compare).
    """

    name = "stablehlo.comparison_direction"


class ComparisonType(StrEnum):
    """
    Together with `comparison_direction` determines the semantics of comparison.

    See StableHLO's [source](https://github.com/openxla/stablehlo/blob/main/stablehlo/dialect/StablehloEnums.td#L152-L156).
    """

    NOTYPE = "NOTYPE"
    FLOAT = "FLOAT"
    TOTALORDER = "TOTALORDER"
    SIGNED = "SIGNED"
    UNSIGNED = "UNSIGNED"


@irdl_attr_definition
class ComparisonTypeAttr(EnumAttribute[ComparisonType], SpacedOpaqueSyntaxAttribute):
    """
    Together with `comparison_direction` determines the semantics of comparison.

    See external [documentation](https://github.com/openxla/stablehlo/blob/main/docs/spec.md#compare).
    """

    name = "stablehlo.comparison_type"


class Precision(StrEnum):
    """
    XLA precision for an operand. Has backend specific meaning.
    """

    DEFAULT = "DEFAULT"
    HIGH = "HIGH"
    HIGHEST = "HIGHEST"


@irdl_attr_definition
class PrecisionAttr(EnumAttribute[Precision], SpacedOpaqueSyntaxAttribute):
    """
    XLA precision for an operand. Has backend specific meaning.

    See external [documentation](https://github.com/openxla/stablehlo/blob/b075e948092d8a27ed0be48f4f8dbaa6df7e2e3e/stablehlo/dialect/StablehloEnums.td#L46).
    """

    name = "stablehlo.precision"


@irdl_attr_definition
class TokenType(TypeAttribute, ParametrizedAttribute):
    """
    Token types represent tokens, i.e. opaque values produced and consumed by some operations.
    Tokens are used for imposing execution order on operations as described in the Execution section.

    E.g.,

    ```mlir
      // %input0: !stablehlo.token
      // %input1: !stablehlo.token
      %result = "stablehlo.after_all"(%input0, %input1) : (!stablehlo.token, !stablehlo.token) -> !stablehlo.token
    ```
    """

    name = "stablehlo.token"


@irdl_attr_definition
class DotAttr(ParametrizedAttribute):
    """
    Attribute that models the dimension information for dot.

    See external [documentation](https://github.com/openxla/stablehlo/blob/b075e948092d8a27ed0be48f4f8dbaa6df7e2e3e/stablehlo/dialect/StablehloAttrs.td#L82).
    """

    name = "stablehlo.dot"

    lhs_batching_dimensions: ParameterDef[ArrayAttr[IntegerAttr[I64]]]
    rhs_batching_dimensions: ParameterDef[ArrayAttr[IntegerAttr[I64]]]
    lhs_contracting_dimensions: ParameterDef[ArrayAttr[IntegerAttr[I64]]]
    rhs_contracting_dimensions: ParameterDef[ArrayAttr[IntegerAttr[I64]]]

    @staticmethod
    def _print_parameter(
        name: str, value: ArrayAttr[IntegerAttr[I64]], printer: Printer
    ):
        printer.print_string(f"\n{name} = [")
        printer.print_list(
            value.data,
            lambda dim: printer.print_string(f"{dim.value.data}"),
        )
        printer.print_string("]")

    @staticmethod
    def _parse_parameter(name: str, parser: AttrParser) -> ArrayAttr[IntegerAttr[I64]]:
        parser.parse_characters(name)
        parser.parse_punctuation("=")
        value = parser.parse_comma_separated_list(
            AttrParser.Delimiter.SQUARE,
            lambda: IntegerAttr(parser.parse_integer(), i64),
        )
        return ArrayAttr(value)

    def print_parameters(self, printer: Printer) -> None:
        with printer.in_angle_brackets():
            with printer.indented():
                DotAttr._print_parameter(
                    "lhs_batching_dimensions", self.lhs_batching_dimensions, printer
                )
                printer.print_string(",")
                DotAttr._print_parameter(
                    "rhs_batching_dimensions", self.rhs_batching_dimensions, printer
                )
                printer.print_string(",")
                DotAttr._print_parameter(
                    "lhs_contracting_dimensions",
                    self.lhs_contracting_dimensions,
                    printer,
                )
                printer.print_string(",")
                DotAttr._print_parameter(
                    "rhs_contracting_dimensions",
                    self.rhs_contracting_dimensions,
                    printer,
                )
            printer.print_string("\n")

    @classmethod
    def parse_parameters(cls, parser: AttrParser) -> Sequence[Attribute]:
        with parser.in_angle_brackets():
            lhs_batching_dimensions = DotAttr._parse_parameter(
                "lhs_batching_dimensions", parser
            )
            parser.parse_punctuation(",")
            rhs_batching_dimensions = DotAttr._parse_parameter(
                "rhs_batching_dimensions", parser
            )
            parser.parse_punctuation(",")
            lhs_contracting_dimensions = DotAttr._parse_parameter(
                "lhs_contracting_dimensions", parser
            )
            parser.parse_punctuation(",")
            rhs_contracting_dimensions = DotAttr._parse_parameter(
                "rhs_contracting_dimensions", parser
            )

            return (
                lhs_batching_dimensions,
                rhs_batching_dimensions,
                lhs_contracting_dimensions,
                rhs_contracting_dimensions,
            )


# endregion


@irdl_op_definition
class AbsOp(IRDLOperation):
    """
    Performs element-wise abs operation on operand tensor and produces a result tensor.
    Depending on the element type, does the following:

    * For signed integers: integer modulus.
    * For floats: abs from IEEE-754.
    * For complex numbers: complex modulus.
    * For quantized types: dequantize_op_quantize(abs, operand, type(result)).

    [See StableHLO specification](https://github.com/openxla/stablehlo/blob/main/docs/spec.md#abs)
    """

    name = "stablehlo.abs"

    # TODO: Remove this constraint for complex types.
    T: ClassVar = VarConstraint("T", base(AnyTensorType))

    operand = operand_def(T)
    result = result_def(T)

    def __init__(self, operand: SSAValue, result_type: Attribute | None = None):
        if result_type is None:
            # TODO: Constraints for complex types.
            result_type = operand.type
        super().__init__(operands=(operand,), result_types=(result_type,))


@irdl_op_definition
class AddOp(ElementwiseBinaryOperation):
    """
    Performs element-wise addition of two tensors `lhs` and `rhs` and produces a
    `result` tensor. Depending on the element type, does the following:

    * For booleans: logical OR.
    * For integers: integer addition.
    * For floats: `addition` from IEEE-754.
    * For complex numbers: complex addition.
    * For quantized types: `dequantize_op_quantize(add, lhs, rhs, type(result))`.

    [See StableHLO specification](https://github.com/openxla/stablehlo/blob/main/docs/spec.md#add)
    """

    name = "stablehlo.add"


@irdl_op_definition
class AfterAllOp(IRDLOperation):
    """
    Ensures that the operations producing the inputs are executed before any operations that depend on result.
    Execution of this operation does nothing, it only exists to establish data dependencies from result to inputs.

    [See StableHLO specification](https://github.com/openxla/stablehlo/blob/main/docs/spec.md#after_all)
    """

    name = "stablehlo.after_all"
    inputs = var_operand_def(TokenType)
    result = result_def(TokenType)

    def __init__(self, inputs: Sequence[SSAValue]):
        super().__init__(operands=[inputs], result_types=(TokenType(),))


@irdl_op_definition
class AndOp(IntegerTensorLikeElementwiseBinaryOperation):
    """
    Performs element-wise AND of two tensors lhs and rhs and produces a result tensor.
    Depending on the element type, does the following:

    For booleans: logical AND.
    For integers: bitwise AND.

    [See StableHLO specification](https://github.com/openxla/stablehlo/blob/main/docs/spec.md#and)
    """

    name = "stablehlo.and"


@irdl_op_definition
class Atan2Op(FloatOrComplexTensorLikeElementwiseBinaryOperation):
    """
    Performs element-wise atan2 operation on `lhs` and `rhs` tensor and produces a
    `result` tensor. Depending on the element type, does the following:

    * For floats: `atan2` from IEEE-754.
    * For complex numbers: complex atan2.
    * For quantized types: `dequantize_op_quantize(atan2, lhs, rhs, type(result))`.

    [See StableHLO specification](https://github.com/openxla/stablehlo/blob/main/docs/spec.md#atan2)
    """

    name = "stablehlo.atan2"


@irdl_op_definition
class BitcastConvertOp(IRDLOperation):
    """
    Performs a bitcast operation on operand tensor and produces a result tensor
    where the bits of the entire operand tensor are reinterpreted using the type of the result tensor.

    More formally, given `E = element_type(operand)`, `E' = element_type(result)`,
    and `R = rank(operand)`:

    * If `num_bits(E') < num_bits(E)`, `bits(result[i0, ..., iR-1, :]) = bits(operand[i0, ..., iR-1])`.
    * If `num_bits(E') > num_bits(E)`, `bits(result[i0, ..., iR-2]) = bits(operand[i0, ..., iR-2, :])`.
    * If `num_bits(E') = num_bits(E)`, `bits(result[i0, ..., iR-1]) = bits(operand[i0, ..., iR-1])`.

    `bits` returns in-memory representation of a given value,
    and its behavior is implementation-defined because the exact representation of tensors is implementation-defined,
    and the exact representation of element types is implementation-defined as well.

    [See StableHLO specification](https://github.com/openxla/stablehlo/blob/main/docs/spec.md#bitcast_convert)
    """

    name = "stablehlo.bitcast_convert"
    input = operand_def(AnyTensorType)
    result = result_def(AnyTensorType)

    def __init__(self, input: SSAValue, result: Attribute):
        super().__init__(operands=(input,), result_types=(result,))


@irdl_op_definition
class CaseOp(IRDLOperation):
    """
    Produces the output from executing exactly one function from `branches`
    depending on the value of `index`. More formally, `result = selected_branch()`
    where:

    * `selected_branch = branches[index]` if `0 <= index < size(branches)`.
    * `selected_branch = branches[-1]` otherwise.

    See [StableHLO specification](https://github.com/openxla/stablehlo/blob/main/docs/spec.md#case)
    """

    name = "stablehlo.case"
    index = operand_def(SI32TensorType)
    branches = var_region_def("single_block")
    _results = var_result_def(AnyTensorTypeConstr | BaseAttr(TokenType))

    def __init__(
        self,
        index: SSAValue,
        branches: Sequence[Region],
        result_types: Sequence[AnyTensorType | TokenType],
    ):
        super().__init__(
            operands=(index,), result_types=(result_types,), regions=(branches,)
        )


@irdl_op_definition
class CbrtOp(FloatOrComplexTensorLikeElementwiseUnaryOperation):
    """
    Performs element-wise cubic root operation on `operand` tensor and produces a
    `result` tensor. Depending on the element type, does the following:

    * For floats: `rootn(x, 3)` from IEEE-754.
    * For complex numbers: complex cubic root.
    * For quantized types: `dequantize_op_quantize(cbrt, operand, type(result))`

    See [StableHLO specification](https://github.com/openxla/stablehlo/blob/main/docs/spec.md#cbrt)
    """

    name = "stablehlo.cbrt"


@irdl_op_definition
class CeilOp(FloatTensorLikeElementwiseUnaryOperation):
    """
    Performs element-wise ceil of operand tensor and produces a result tensor.
    Implements the roundToIntegralTowardPositive operation from the IEEE-754 specification.
    For quantized types, performs dequantize_op_quantize(ceil, operand, type(result)).

    See [StableHLO specification](https://github.com/openxla/stablehlo/blob/main/docs/spec.md#ceil)
    """

    name = "stablehlo.ceil"


@irdl_op_definition
class CountLeadingZerosOp(IntegerTensorLikeElementwiseUnaryOperation):
    """
    Performs element-wise ceil of `operand` tensor and produces a `result` tensor.
    Implements the `roundToIntegralTowardPositive` operation from the IEEE-754
    specification. For quantized types, performs
    `dequantize_op_quantize(ceil, operand, type(result))`.

    See [StableHLO specification](https://github.com/openxla/stablehlo/blob/main/docs/spec.md#count_leading_zeros)
    """

    name = "stablehlo.count_leading_zeros"


@irdl_op_definition
class MultiplyOp(ElementwiseBinaryOperation):
    """
    Performs element-wise product of two tensors `lhs` and `rhs` and produces a
    `result` tensor. Depending on the element type, does the following:

    * For booleans: logical AND.
    * For integers: integer multiplication.
    * For floats: `multiplication` from IEEE-754.
    * For complex numbers: complex multiplication.
    * For quantized types:
      * `dequantize_op_quantize(multiply, lhs, rhs, type(result))`.

    See [StableHLO specification](https://github.com/openxla/stablehlo/blob/main/docs/spec.md#multiply)
    """

    name = "stablehlo.multiply"


@irdl_op_definition
class NotOp(IntegerTensorLikeElementwiseUnaryOperation):
    """
    Performs element-wise NOT of tensor `operand` and produces a `result` tensor.
    Depending on the element type, does the following:

    * For booleans: logical NOT.
    * For integers: bitwise NOT.

    See [StableHLO specification](https://github.com/openxla/stablehlo/blob/main/docs/spec.md#not)
    """

    name = "stablehlo.not"


@irdl_op_definition
class OrOp(IntegerTensorLikeElementwiseBinaryOperation):
    """
    Performs element-wise OR of two tensors `lhs` and `rhs` and produces a `result`
    tensor. Depending on the element type, does the following:

    * For booleans: logical OR.
    * For integers: bitwise OR.

    See [StableHLO specification](https://github.com/openxla/stablehlo/blob/main/docs/spec.md#or)
    """

    name = "stablehlo.or"


@irdl_op_definition
class PopcntOp(IntegerTensorLikeElementwiseUnaryOperation):
    """
    Performs element-wise count of the number of bits set in the `operand` tensor
    and produces a `result` tensor.

    See [StableHLO specification](https://github.com/openxla/stablehlo/blob/main/docs/spec.md#popcnt)
    """

    name = "stablehlo.popcnt"


@irdl_op_definition
class ReturnOp(IRDLOperation):
    """This op is un-documented.

    StableHLO's return is used inside of the bodies of StableHLO ops.
    It behaves like func.return but for StableHLO ops.
    The func.return op is used inside of func.func op.

    https://discord.com/channels/999073994483433573/1259494021269688360/1259992088565645312
    """

    name = "stablehlo.return"

    input = var_operand_def(AnyTensorType)
    traits = traits_def(IsTerminator())

    def __init__(self, input: list[SSAValue]):
        super().__init__(operands=(input,))


@irdl_op_definition
class ShiftLeftOp(IntegerTensorLikeElementwiseBinaryOperation):
    """
    Performs element-wise left-shift operation on the `lhs` tensor by `rhs` number
    of bits and produces a `result` tensor.

    See [StableHLO specification](https://github.com/openxla/stablehlo/blob/main/docs/spec.md#shift_left)
    """

    name = "stablehlo.shift_left"


@irdl_op_definition
class ShiftRightArithmeticOp(IntegerTensorLikeElementwiseBinaryOperation):
    """
    Performs element-wise arithmetic right-shift operation on the `lhs` tensor by
    `rhs` number of bits and produces a `result` tensor.

    See [StableHLO specification](https://github.com/openxla/stablehlo/blob/main/docs/spec.md#shift_right_arithmetic)
    """

    name = "stablehlo.shift_right_arithmetic"


@irdl_op_definition
class ShiftRightLogicalOp(IntegerTensorLikeElementwiseBinaryOperation):
    """
    Performs element-wise logical right-shift operation on the `lhs` tensor by `rhs`
    number of bits and produces a `result` tensor.

    See [StableHLO specification](https://github.com/openxla/stablehlo/blob/main/docs/spec.md#shift_right_logical)
    """

    name = "stablehlo.shift_right_logical"


@irdl_op_definition
class SubtractOp(ElementwiseBinaryOperation):
    """
    Performs element-wise subtraction of two tensors `lhs` and `rhs` and produces a
    `result` tensor. Depending on the element type, does the following:

    * For integers: integer subtraction.
    * For floats: `subtraction` from IEEE-754.
    * For complex numbers: complex subtraction.
    * For quantized types:
      * `dequantize_op_quantize(subtract, lhs, rhs, type(result))`.

    See [StableHLO specification](https://github.com/openxla/stablehlo/blob/main/docs/spec.md#subtract)
    """

    name = "stablehlo.subtract"


@irdl_op_definition
class TransposeOp(IRDLOperation):
    """
    Permutes the dimensions of `operand` tensor using `permutation` and produces a
    `result` tensor. More formally, `result[result_index] = operand[operand_index]`
    where `result_index[d] = operand_index[permutation[d]]`.

    See [StableHLO specification](https://github.com/openxla/stablehlo/blob/main/docs/spec.md#transpose)
    """

    name = "stablehlo.transpose"

    ElementType = Annotated[Attribute, ConstraintVar("ElementType")]

    operand = operand_def(TensorType[ElementType])
    result = result_def(TensorType[ElementType])
    permutation = attr_def(DenseArrayBase)

    def __init__(
        self, operand: SSAValue, permutation: DenseArrayBase, result_type: Attribute
    ):
        super().__init__(
            operands=(operand,),
            result_types=(result_type,),
            attributes={"permutation": permutation},
        )

    def get_permutation(self) -> tuple[int, ...]:
        return cast(tuple[int, ...], self.permutation.get_values())

    def verify_(self) -> None:
        # Operand and result types are checked before the custom `verify_`
        o_type = cast(TensorType[Attribute], self.operand.type)
        r_type = cast(TensorType[Attribute], self.result.type)

        o_shape = o_type.get_shape()
        r_shape = r_type.get_shape()

        # TODO: Quantization constraints
        # `permutation` is a permutation of `range(rank(operand))`
        permutation = self.get_permutation()
        if sorted(permutation) != list(range(len(o_shape))):
            raise VerifyException(
                f"Permutation {permutation} of transpose must be a permutation of "
                f"range({len(o_shape)})"
            )

        # `shape(result) = dim(operand, permutation...)`
        for i, dim in enumerate(permutation):
            if r_shape[i] != o_shape[dim]:
                raise VerifyException(
                    f"Permutation mismatch at dimension {i}, expected {o_shape[dim]}"
                )


@irdl_op_definition
class XorOp(IntegerTensorLikeElementwiseBinaryOperation):
    """
    Performs element-wise XOR of two tensors `lhs` and `rhs` and produces a `result`
    tensor. Depending on the element type, does the following:

    * For booleans: logical XOR.
    * For integers: bitwise XOR.

    [See StableHLO specification](https://github.com/openxla/stablehlo/blob/main/docs/spec.md#xor)
    """

    name = "stablehlo.xor"


StableHLO = Dialect(
    "stablehlo",
    [
        AbsOp,
        AddOp,
        AfterAllOp,
        AndOp,
        Atan2Op,
        BitcastConvertOp,
        CaseOp,
        CbrtOp,
        CeilOp,
        CountLeadingZerosOp,
        MultiplyOp,
        NotOp,
        OrOp,
        PopcntOp,
        ReturnOp,
        ShiftLeftOp,
        ShiftRightArithmeticOp,
        ShiftRightLogicalOp,
        SubtractOp,
        TransposeOp,
        XorOp,
    ],
    [
        ComparisonDirectionAttr,
        ComparisonTypeAttr,
        DotAttr,
        PrecisionAttr,
        TokenType,
    ],
)



from __future__ import annotations

from xdsl.dialects.arm.assembly import AssemblyInstructionArg
from xdsl.dialects.arm.register import ARMRegisterType
from xdsl.ir import (
    Dialect,
    EnumAttribute,
    SpacedOpaqueSyntaxAttribute,
    SSAValue,
    StrEnum,
)
from xdsl.irdl import (
    irdl_attr_definition,
)

ARM_NEON_INDEX_BY_NAME = {f"v{i}": i for i in range(0, 32)}


@irdl_attr_definition
class NEONRegisterType(ARMRegisterType):
    """
    A 128-bit NEON ARM register type.
    """

    name = "arm_neon.reg"

    @classmethod
    def instruction_set_name(cls) -> str:
        return "arm_neon"

    @classmethod
    def index_by_name(cls) -> dict[str, int]:
        return ARM_NEON_INDEX_BY_NAME

    @classmethod
    def infinite_register_prefix(cls):
        return "inf_"


UNALLOCATED_NEON = NEONRegisterType.unallocated()
V0 = NEONRegisterType.from_name("v0")
V1 = NEONRegisterType.from_name("v1")
V2 = NEONRegisterType.from_name("v2")
V3 = NEONRegisterType.from_name("v3")
V4 = NEONRegisterType.from_name("v4")
V5 = NEONRegisterType.from_name("v5")
V6 = NEONRegisterType.from_name("v6")
V7 = NEONRegisterType.from_name("v7")
V8 = NEONRegisterType.from_name("v8")
V9 = NEONRegisterType.from_name("v9")
V10 = NEONRegisterType.from_name("v10")
V11 = NEONRegisterType.from_name("v11")
V12 = NEONRegisterType.from_name("v12")
V13 = NEONRegisterType.from_name("v13")
V14 = NEONRegisterType.from_name("v14")
V15 = NEONRegisterType.from_name("v15")
V16 = NEONRegisterType.from_name("v16")
V17 = NEONRegisterType.from_name("v17")
V18 = NEONRegisterType.from_name("v18")
V19 = NEONRegisterType.from_name("v19")
V20 = NEONRegisterType.from_name("v20")
V21 = NEONRegisterType.from_name("v21")
V22 = NEONRegisterType.from_name("v22")
V23 = NEONRegisterType.from_name("v23")
V24 = NEONRegisterType.from_name("v24")
V25 = NEONRegisterType.from_name("v25")
V26 = NEONRegisterType.from_name("v26")
V27 = NEONRegisterType.from_name("v27")
V28 = NEONRegisterType.from_name("v28")
V29 = NEONRegisterType.from_name("v29")
V30 = NEONRegisterType.from_name("v30")
V31 = NEONRegisterType.from_name("v31")


class NeonArrangement(StrEnum):
    """
    The arrangement specifier for NEON instructions determines element size and count.
    We assume full 128-bit registers. Possible arrangements:
      - D  → 2 double-precision floats
      - S  → 4 single-precision floats
      - H  → 8 half-precision floats
    """

    D = "D"
    S = "S"
    H = "H"

    def map_to_num_els(self):
        map = {"D": 2, "S": 4, "H": 8}
        return map[self.name]


@irdl_attr_definition
class NeonArrangementAttr(EnumAttribute[NeonArrangement], SpacedOpaqueSyntaxAttribute):
    """
    Attribute containing the arrangement specification.
    """

    name = "arm_neon.arrangement"


class VectorWithArrangement(AssemblyInstructionArg):
    reg: NEONRegisterType | SSAValue
    arrangement: NeonArrangementAttr
    index: int | None = None

    def __init__(
        self,
        reg: NEONRegisterType,
        arrangement: NeonArrangementAttr,
        *,
        index: int | None = None,
    ):
        self.reg = reg
        self.arrangement = arrangement
        self.index = index

    def assembly_str(self):
        assert isinstance(self.reg, NEONRegisterType)
        if self.index is None:
            return f"{self.reg.register_name.data}.{self.arrangement.data.map_to_num_els()}{self.arrangement.data.name}"
        else:
            return f"{self.reg.register_name.data}.{self.arrangement.data.name}[{self.index}]"


ARM_NEON = Dialect(
    "arm_neon",
    [],
    [
        NeonArrangementAttr,
        NEONRegisterType,
    ],
)



from io import StringIO

import pytest

from xdsl.context import Context
from xdsl.dialects.builtin import DYNAMIC_INDEX, IndexType, IntegerType, i32
from xdsl.dialects.utils import (
    parse_dynamic_index_list_with_types,
    parse_dynamic_index_list_without_types,
    parse_dynamic_index_with_type,
    parse_dynamic_index_without_type,
    print_dynamic_index_list,
)
from xdsl.ir import Dialect, SSAValue
from xdsl.parser import Parser, UnresolvedOperand
from xdsl.printer import Printer
from xdsl.utils.test_value import TestSSAValue

ctx = Context()
index = IndexType()


def test_print_dynamic_index_list():
    # Test case 1: Only integers
    stream = StringIO()
    printer = Printer(stream)
    print_dynamic_index_list(printer, [], [1, 2, 3])
    assert stream.getvalue() == "[1, 2, 3]"

    # Test case 2: Mix of integers and SSA values
    stream = StringIO()
    printer = Printer(stream)
    values = [TestSSAValue(IndexType()), TestSSAValue(IndexType())]
    print_dynamic_index_list(printer, values, [1, DYNAMIC_INDEX, 3, DYNAMIC_INDEX])
    assert stream.getvalue() == "[1, %0, 3, %1]"

    # Test case 3: With value types
    stream = StringIO()
    printer = Printer(stream)
    values = [TestSSAValue(IndexType()), TestSSAValue(IntegerType(32))]
    value_types = (IndexType(), IntegerType(32))
    print_dynamic_index_list(
        printer, values, [DYNAMIC_INDEX, 2, DYNAMIC_INDEX], value_types
    )
    assert stream.getvalue() == "[%0 : index, 2, %1 : i32]"

    # Test case 4: Custom delimiter
    stream = StringIO()
    printer = Printer(stream)
    print_dynamic_index_list(printer, [], [1, 2, 3], delimiter=Parser.Delimiter.PAREN)
    assert stream.getvalue() == "(1, 2, 3)"

    # Test case 5: Empty list
    stream = StringIO()
    printer = Printer(stream)
    print_dynamic_index_list(printer, [], [])
    assert stream.getvalue() == "[]"

    # Test case 6: Mix of integers and SSA values with custom dynamic index
    dynamic_index = -42
    stream = StringIO()
    printer = Printer(stream)
    values = [TestSSAValue(IndexType()), TestSSAValue(IndexType())]
    print_dynamic_index_list(
        printer,
        values,
        [1, dynamic_index, 3, dynamic_index],
        dynamic_index=dynamic_index,
    )
    assert stream.getvalue() == "[1, %0, 3, %1]"


@pytest.mark.parametrize(
    "delimiter,expected",
    [
        (Parser.Delimiter.SQUARE, "[1, 2, 3]"),
        (Parser.Delimiter.PAREN, "(1, 2, 3)"),
        (Parser.Delimiter.ANGLE, "<1, 2, 3>"),
        (Parser.Delimiter.BRACES, "{1, 2, 3}"),
    ],
)
def test_print_dynamic_index_list_delimiters(
    delimiter: Parser.Delimiter, expected: str
):
    stream = StringIO()
    printer = Printer(stream)
    print_dynamic_index_list(printer, [], [1, 2, 3], delimiter=delimiter)
    assert stream.getvalue() == expected


def test_parse_dynamic_index_with_type():
    parser = Parser(ctx, "%0 : i32")
    result = parse_dynamic_index_with_type(parser)
    assert isinstance(result, SSAValue)

    parser = Parser(ctx, "42")
    result = parse_dynamic_index_with_type(parser)
    assert isinstance(result, int)
    assert result == 42


def test_parse_dynamic_index_without_type():
    parser = Parser(ctx, "%0")
    result = parse_dynamic_index_without_type(parser)
    assert isinstance(result, UnresolvedOperand)

    parser = Parser(ctx, "42")
    result = parse_dynamic_index_without_type(parser)
    assert isinstance(result, int)
    assert result == 42


def test_parse_dynamic_index_list_with_types():
    dynamic_index = -42
    test_values = (TestSSAValue(i32), TestSSAValue(index))
    parser = Parser(ctx, "[%0 : i32, 42, %1 : index]")
    parser.ssa_values["0"] = (test_values[0],)
    parser.ssa_values["1"] = (test_values[1],)
    values, indices = parse_dynamic_index_list_with_types(
        parser, dynamic_index=dynamic_index
    )
    assert len(values) == 2
    assert values[0] is test_values[0]
    assert values[1] is test_values[1]
    assert tuple(indices) == (dynamic_index, 42, dynamic_index)


def test_parse_dynamic_index_list_without_types():
    dynamic_index = -42
    parser = Parser(ctx, "[%0, 42, %1]")
    values, indices = parse_dynamic_index_list_without_types(
        parser, dynamic_index=dynamic_index
    )

    assert len(values) == 2
    assert isinstance(values[0], UnresolvedOperand)
    assert isinstance(values[1], UnresolvedOperand)

    assert tuple(indices) == (dynamic_index, 42, dynamic_index)


def test_parse_dynamic_index_list_with_custom_delimiter():
    dynamic_index = -42
    test_values = (TestSSAValue(i32), TestSSAValue(index))
    parser = Parser(ctx, "(%0 : i32, 42, %1 : index)")
    parser.ssa_values["0"] = (test_values[0],)
    parser.ssa_values["1"] = (test_values[1],)
    values, indices = parse_dynamic_index_list_with_types(
        parser, dynamic_index=dynamic_index, delimiter=Parser.Delimiter.PAREN
    )
    assert len(values) == 2
    assert values[0] is test_values[0]
    assert values[1] is test_values[1]
    assert tuple(indices) == (dynamic_index, 42, dynamic_index)


@pytest.mark.parametrize(
    "name,expected_1,expected_2",
    [
        ("dialect.op_name", "dialect", "op_name"),
        ("dialect.op.name", "dialect", "op.name"),
    ],
)
def test_split_name(name: str, expected_1: str, expected_2: str):
    result_1, result_2 = Dialect.split_name(name)
    assert result_1 == expected_1
    assert result_2 == expected_2


def test_split_name_failure():
    with pytest.raises(ValueError) as e:
        Dialect.split_name("test")

    assert e.value.args[0] == ("Invalid operation or attribute name test.")



import types
from collections.abc import Iterable, Sequence
from inspect import isclass
from typing import (
    TYPE_CHECKING,
    Annotated,
    Any,
    Generic,
    Literal,
    TypeGuard,
    TypeVar,
    Union,
    cast,
    get_args,
    get_origin,
)

from xdsl.ir import ParametrizedAttribute
from xdsl.utils.exceptions import VerifyException

if TYPE_CHECKING:
    from typing_extensions import TypeForm

_T = TypeVar("_T")


def isa(arg: Any, hint: "TypeForm[_T]") -> TypeGuard[_T]:
    from xdsl.irdl import ConstraintContext

    """
    Check if `arg` is of the type described by `hint`.
    For now, only lists, dictionaries, unions,
    and non-generic classes are supported for type hints.
    """
    if hint is Any:  # pyright: ignore[reportUnnecessaryComparison]
        return True

    origin = get_origin(hint)

    # get_origin checks that hint is not a parametrized generic
    if isclass(hint) and (origin is None):
        return isinstance(arg, hint)

    if origin is list:
        if not isinstance(arg, list):
            return False
        arg_list: list[Any] = cast(list[Any], arg)
        (elem_hint,) = get_args(hint)
        return all(isa(elem, elem_hint) for elem in arg_list)

    if origin is set:
        if not isinstance(arg, set):
            return False
        arg_set: set[Any] = cast(set[Any], arg)
        (elem_hint,) = get_args(hint)
        return all(isa(elem, elem_hint) for elem in arg_set)

    if origin is tuple:
        if not isinstance(arg, tuple):
            return False
        elem_hints = get_args(hint)
        arg_tuple: tuple[Any, ...] = cast(tuple[Any, ...], arg)
        if len(elem_hints) == 2 and elem_hints[1] is ...:
            return all(isa(elem, elem_hints[0]) for elem in arg_tuple)
        else:
            return len(elem_hints) == len(arg_tuple) and all(
                isa(elem, hint) for elem, hint in zip(arg_tuple, elem_hints)
            )

    if origin is dict:
        if not isinstance(arg, dict):
            return False
        arg_dict: dict[Any, Any] = cast(dict[Any, Any], arg)
        key_hint, value_hint = get_args(hint)
        return all(
            isa(key, key_hint) and isa(value, value_hint)
            for key, value in arg_dict.items()
        )

    if origin in [Union, types.UnionType]:
        return any(isa(arg, union_arg) for union_arg in get_args(hint))

    if origin is Literal:
        return arg in get_args(hint)

    if (origin is not None) and issubclass(origin, Sequence):
        if not isinstance(arg, Sequence):
            return False
        arg_seq = cast(Sequence[Any], arg)
        (elem_hint,) = get_args(hint)
        return all(isa(elem, elem_hint) for elem in arg_seq)

    from xdsl.irdl import GenericData, irdl_to_attr_constraint

    if (origin is not None) and issubclass(origin, GenericData | ParametrizedAttribute):
        constraint = irdl_to_attr_constraint(
            hint  # pyright: ignore[reportArgumentType]
        )
        try:
            constraint.verify(arg, ConstraintContext())
            return True
        except VerifyException:
            return False

    raise ValueError(f"isa: unsupported type hint '{hint}' {get_origin(hint)}")


def assert_isa(arg: Any, hint: "TypeForm[_T]") -> TypeGuard[_T]:
    """
    Check if `arg` is of the type described by `hint`.
    For now, only lists, dictionaries, unions,
    and non-generic classes are supported for type hints.
    """

    if not isa(arg, hint):
        raise ValueError(
            f"Expected value of type {hint}, got value of type {type(arg).__name__}"
        )
    return True


annotated_type = type(Annotated[int, 0])
"""This is the type of an Annotated object."""


class _Class:
    @property
    def property(self):
        pass


PropertyType = type(_Class.property)
"""The type of a property method."""


def get_type_var_from_generic_class(cls: type[Any]) -> tuple[TypeVar, ...]:
    """Return the `TypeVar` used in the class `Generic` parent."""
    cls_orig_bases: Iterable[Any] = getattr(cls, "__orig_bases__")
    for orig_base in cls_orig_bases:
        if get_origin(orig_base) == Generic:
            return get_args(orig_base)
    raise ValueError(f"{cls} class does not have a `Generic` parent.")


def get_type_var_mapping(
    cls: type[Any],
) -> tuple[type[Any], dict[TypeVar, Any]]:
    """
    Given a class that specializes a generic class, return the generic class and
    the mapping from the generic class type variables to the specialized arguments.
    """

    if not issubclass(cls, Generic):
        raise ValueError(f"{cls} does not specialize a generic class.")

    # Get the generic parent
    orig_bases: Iterable[Any] = getattr(cls, "__orig_bases__")
    orig_bases = [
        orig_base
        for orig_base in orig_bases
        if (origin := get_origin(orig_base)) is not Generic and origin is not None
    ]
    # Do not handle more than one generic parent in the mro.
    # It is possible to handle more than one generic parent, but
    # the mapping of type variables will be more complex, especially for
    # generic parents inheriting from other generic parents.
    if len(orig_bases) != 1:
        raise ValueError(
            "Class cannot have more than one generic class in its mro. This "
            "restriction may be lifted in the future.",
            orig_bases,
        )

    # Get the generic operation, and its specialized type parameters.
    generic_parent: type[Any] = get_origin(orig_bases[0])
    specialized_args = get_args(orig_bases[0])

    # Get the `TypeVar` used in the generic parent
    generic_args = get_type_var_from_generic_class(generic_parent)

    if len(generic_args) != len(specialized_args):
        raise ValueError(
            f"Generic class {generic_parent} class has {len(generic_args)} "
            f"parameters, but {cls} specialize {len(specialized_args)} of them."
        )

    type_var_mapping = dict(zip(generic_args, specialized_args))
    return generic_parent, type_var_mapping


def type_repr(obj: Any) -> str:
    """Return the repr() of an object, special-casing types."""
    if isinstance(obj, types.GenericAlias):
        origin = get_origin(obj)
        args = get_args(obj)
        return f"{type_repr(origin)}[{', '.join(type_repr(arg) for arg in args)}]"
    if isinstance(obj, types.UnionType):
        args = get_args(obj)
        return f"{'|'.join(type_repr(arg) for arg in args)}"
    if obj is type(None):
        return "None"
    if obj is ...:
        return "..."
    if isinstance(obj, types.FunctionType):
        return obj.__name__
    if isinstance(obj, type):
        return f"{obj.__qualname__}"
    return repr(obj)



import marimo

__generated_with = "0.11.19"
app = marimo.App(width="medium")


@app.cell(hide_code=True)
def _():
    import marimo as mo
    import subprocess
    return mo, subprocess


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""
        # xDSL-MLIR Interoperation Tutorial

        This tutorial aims to showcase a simple pipeline of actions to unlock MLIR optimisations when lowering from xDSL.
        This tutorial can help users getting familiar with the xDSL-MLIR interoperation. We will start from a higher level of xDSL abstraction, lower to MLIR generic format, apply an optimisation and the return to xDSL-land.
        """
    )
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""
        ## Problem Setup

        We start by writing a simple example consisting of adding integers.
        We are writing this example using constructs that are supported in xDSL.

        Furthermore, we create 4 integers, namely a, b, c, d.
        Then we just accumulate by the simple following pseudocode and print the result:


        ```bash
        a = 1
        b = 2
        c = a + b
        d = a + b
        e = c + d
        print(e)
        ```
        """
    )
    return


@app.cell
def _():
    from xdsl.dialects.arith import AddiOp, ConstantOp
    from xdsl.dialects.builtin import IntegerAttr, ModuleOp, i32
    from xdsl.dialects.vector import PrintOp
    from xdsl.ir import Block, Region

    # Define two integer constants
    a = ConstantOp(IntegerAttr.from_int_and_width(1, 32), i32)
    b = ConstantOp(IntegerAttr.from_int_and_width(2, 32), i32)

    # Operations on these constants
    c = AddiOp(a, b)
    d = AddiOp(a, b)
    e = AddiOp(c, d)
    f = PrintOp.get(e)

    # Create Block from operations and Region from blocks
    block0 = Block([a, b, c, d, e, f])
    region0 = Region(block0)

    # Create an Operation from the region
    op = ModuleOp(region0)
    return (
        AddiOp,
        Block,
        ConstantOp,
        IntegerAttr,
        ModuleOp,
        PrintOp,
        Region,
        a,
        b,
        block0,
        c,
        d,
        e,
        f,
        i32,
        op,
        region0,
    )


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""
        Using xDSLs printer we can print this operation.
        For convenience we provide a file called `source.mlir` with the code printed below
        """
    )
    return


@app.cell
def _(op):
    from xdsl.printer import Printer

    # Print in xdsl format
    printer = Printer()
    printer.print(op)
    return Printer, printer


@app.cell
def _(mo, subprocess):
    # Cross-check file content
    source_file = mo.notebook_dir() / "source.mlir"
    subprocess.run(["cat", source_file])
    return (source_file,)


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""
        Now, let's try to benefit from some MLIR optimization.
        For this example, we will use the [Common subexpression elimination](https://en.wikipedia.org/wiki/Common_subexpression_elimination).

        See some documentation here: [mlir.llvm CSE docs](https://mlir.llvm.org/docs/Passes/#-cse-eliminate-common-sub-expressions).

        Assuming you have already `mlir-opt` installed in your machine:
        """
    )
    return


@app.cell(hide_code=True)
def _(subprocess):
    mlir_opt_tool = "mlir-opt"
    is_mlir_opt_available = False

    try:
        subprocess.run([mlir_opt_tool, "--version"], stdout=subprocess.PIPE, stderr=subprocess.PIPE, check=True)
        is_mlir_opt_available = True
    except (subprocess.CalledProcessError, FileNotFoundError):
        pass

    if is_mlir_opt_available:
        print(f"{mlir_opt_tool} is available.")
    else:
        print(f"{mlir_opt_tool} is not available.")
    return is_mlir_opt_available, mlir_opt_tool


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""you can apply the CSE optimization using the following command:""")
    return


@app.cell
def _(mlir_opt_tool, source_file, subprocess):
    # mlir-opt --cse --mlir-print-op-generic
    ps1 = subprocess.run([mlir_opt_tool, source_file, "--cse", "--mlir-print-op-generic"], capture_output=True)
    print(ps1.stdout.decode('utf-8').strip())
    return (ps1,)


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""
        We can clearly see in the optimized output that after CSE we do not need to calculate:

        ```
        arith.addi"(%0, %1) : (i32, i32) -> i32
        ```

        twice!

        Now can we back to xDSL? Yes we can!
        """
    )
    return


@app.cell
def _(ps1, subprocess):
    subprocess.run(["xdsl-opt"], input=ps1.stdout)
    return


if __name__ == "__main__":
    app.run()



# RUN: python %s | filecheck %s

from xdsl.frontend.pyast.context import CodeContext
from xdsl.frontend.pyast.dialects.builtin import i1, i32, i64
from xdsl.frontend.pyast.exception import (
    CodeGenerationException,
    FrontendProgramException,
)
from xdsl.frontend.pyast.program import FrontendProgram

p = FrontendProgram()

try:
    with CodeContext(p):
        # CHECK: Expected non-zero number of return types in function 'test_no_return_type', but got 0.
        def test_no_return_type(a: i32) -> i32:
            return

    p.compile(desymref=False)
    print(p.textual_format())
except FrontendProgramException as e:
    print(e.msg)

try:
    with CodeContext(p):
        # CHECK: Type signature and the type of the return value do not match at position 0: expected i32, got i64.
        def test_wrong_return_type(a: i32, b: i64) -> i32:
            return b

    p.compile(desymref=False)
    print(p.textual_format())
except FrontendProgramException as e:
    print(e.msg)

try:
    with CodeContext(p):
        # CHECK: Expected no return types in function 'test_wrong_return_type'.
        def test_wrong_return_type(a: i32):
            return a

    p.compile(desymref=False)
    print(p.textual_format())
except FrontendProgramException as e:
    print(e.msg)

try:
    with CodeContext(p):
        # CHECK: Expected the same types for binary operation 'Add', but got i32 and i64.
        def bin_op_type_mismatch(a: i32, b: i64) -> i32:
            return a + b

    p.compile(desymref=False)
    print(p.textual_format())
except FrontendProgramException as e:
    print(e.msg)

try:
    with CodeContext(p):
        # CHECK: Expected the same types for comparison operator 'Lt', but got i32 and i64.
        def cmp_op_type_mismatch(a: i32, b: i64) -> i1:
            return a < b

    p.compile(desymref=False)
    print(p.textual_format())
except FrontendProgramException as e:
    print(e.msg)

try:
    with CodeContext(p):
        # CHECK: Else clause in for loops is not supported.
        def test_not_supported_loop_I():
            for i in range(10):
                pass
            else:
                pass
            return

    p.compile(desymref=False)
    print(p.textual_format())
except CodeGenerationException as e:
    print(e.msg)

try:
    with CodeContext(p):
        # CHECK: Only range-based loops are supported.
        def test_not_supported_loop_II():
            for i, j in enumerate(range(10, 0, -1)):
                pass
            return

    p.compile(desymref=False)
    print(p.textual_format())
except CodeGenerationException as e:
    print(e.msg)

try:
    with CodeContext(p):
        # CHECK: Range-based loop expected between 1 and 3 arguments, but got 4.
        def test_not_supported_loop_III():
            for i in range(0, 1, 2, 3):
                pass
            return

    p.compile(desymref=False)
    print(p.textual_format())
except CodeGenerationException as e:
    print(e.msg)

try:
    with CodeContext(p):
        # CHECK: Range-based loop must take a single target variable, e.g. `for i in range(10)`.
        def test_not_supported_loop_IV():
            for i, j in range(100):
                pass
            return

    p.compile(desymref=False)
    print(p.textual_format())
except CodeGenerationException as e:
    print(e.msg)



import re
from dataclasses import dataclass

from typing_extensions import Any

from xdsl.utils.lexer import Token


@dataclass
class Location:
    "Structure definition a location in a file."

    file: str
    line: int
    col: int

    def __repr__(self):
        return f"{self.file}:{self.line}:{self.col}"


_NEWLINE = re.compile(r"\n")


def loc(token: Token[Any]) -> Location:
    file = token.span.input.name
    # Could be much faster

    remaining = token.span.start
    prev_end = 0

    for line, newline_match in enumerate(
        re.finditer(_NEWLINE, token.span.input.content)
    ):
        len_line = newline_match.start() - prev_end
        if remaining < len_line:
            return Location(file, line + 1, remaining + 1)
        remaining -= len_line + 1
        prev_end = newline_match.end()

    raise AssertionError(f"Could not find location of token {token}")



"""
A dialect for handling distributed memory parallelism (DMP).

This is xDSL only for now.

This dialect aims to provide the tools necessary to facilitate the creation
and lowering of stencil (and other) computations in a manner that
makes them run on node clusters.
"""

from __future__ import annotations

from abc import ABC
from collections.abc import Iterable, Sequence
from math import prod
from typing import Literal, cast

from xdsl.dialects import builtin, stencil
from xdsl.ir import (
    Attribute,
    Dialect,
    Operation,
    ParametrizedAttribute,
    SSAValue,
)
from xdsl.irdl import (
    IRDLOperation,
    ParameterDef,
    attr_def,
    irdl_attr_definition,
    irdl_op_definition,
    operand_def,
    opt_result_def,
    traits_def,
)
from xdsl.parser import AttrParser
from xdsl.printer import Printer
from xdsl.traits import (
    EffectInstance,
    HasShapeInferencePatternsTrait,
    MemoryEffect,
    MemoryEffectKind,
)
from xdsl.utils.exceptions import VerifyException
from xdsl.utils.hints import isa

# helpers for named dimensions:
DIM_X = 0
DIM_Y = 1
DIM_Z = 2


@irdl_attr_definition
class ExchangeDeclarationAttr(ParametrizedAttribute):
    """
    This declares a region to be "halo-exchanged".
    The semantics define that the region specified by offset and size
    is the *received part*. To get the section that should be sent,
    use the source_area() method to get the source area.

     - offset gives the coordinates from the origin of the stencil field.
     - size gives the size of the buffer to be exchanged.
     - source_offset gives a translation (n-d offset) where the data should be
       read from that is exchanged with the other node.
     - neighbor gives the offset in rank to the node this data is to be
       exchanged with

    Example:

        offset = [4, 0]
        size   = [10, 1]
        source_offset = [0, 1]
        neighbor = -1

    To visualize:
    0   4         14
        xxxxxxxxxx    0
        oooooooooo    1

    Where `x` signifies the area that should be received,
    and `o` the area that should be read from.

    This data will be exchanged with the node of rank (my_rank -1)
    """

    name = "dmp.exchange"

    offset_: ParameterDef[builtin.DenseArrayBase]
    size_: ParameterDef[builtin.DenseArrayBase]
    source_offset_: ParameterDef[builtin.DenseArrayBase]
    neighbor_: ParameterDef[builtin.DenseArrayBase]

    def __init__(
        self,
        offset: Sequence[int],
        size: Sequence[int],
        source_offset: Sequence[int],
        neighbor: Sequence[int],
    ):
        data_type = builtin.i64
        super().__init__(
            [
                builtin.DenseArrayBase.from_list(data_type, offset),
                builtin.DenseArrayBase.from_list(data_type, size),
                builtin.DenseArrayBase.from_list(data_type, source_offset),
                builtin.DenseArrayBase.from_list(data_type, neighbor),
            ]
        )

    @classmethod
    def from_points(
        cls,
        points: Sequence[tuple[int, int]],
        dim: int,
        dir_sign: Literal[1, -1],
        neighbor_offset: int = 1,
    ):
        sizes = tuple(e - s for s, e in points)
        return cls(
            # get starting points
            tuple(s for s, _ in points),
            # calculated sizes
            sizes,
            # source_offset (opposite of exchange direction)
            tuple(
                0 if d != dim else -1 * dir_sign * sizes[dim] * neighbor_offset
                for d in range(len(sizes))
            ),
            # direction
            tuple(
                0 if d != dim else dir_sign * neighbor_offset for d in range(len(sizes))
            ),
        )

    @property
    def offset(self) -> tuple[int, ...]:
        data = self.offset_.get_values()
        assert isa(data, tuple[int, ...])
        return data

    @property
    def size(self) -> tuple[int, ...]:
        data = self.size_.get_values()
        assert isa(data, tuple[int, ...])
        return data

    @property
    def source_offset(self) -> tuple[int, ...]:
        data = self.source_offset_.get_values()
        assert isa(data, tuple[int, ...])
        return data

    @property
    def neighbor(self) -> tuple[int, ...]:
        data = self.neighbor_.get_values()
        assert isa(data, tuple[int, ...])
        return data

    @property
    def elem_count(self) -> int:
        return prod(self.size)

    @property
    def dims(self) -> int:
        """
        number of dimensions of the grid
        """
        return len(self.size)

    def source_area(self) -> ExchangeDeclarationAttr:
        """
        Since a HaloExchangeDef by default specifies the area to receive into,
        this method returns the area that should be read from.
        """
        # we set source_offset to all zero, so that repeated calls to source_area never
        # return the dest area
        return ExchangeDeclarationAttr(
            offset=tuple(
                val + offs for val, offs in zip(self.offset, self.source_offset)
            ),
            size=self.size,
            source_offset=tuple(0 for _ in range(len(self.source_offset))),
            neighbor=self.neighbor,
        )

    def print_parameters(self, printer: Printer) -> None:
        printer.print_string("<at [")
        printer.print_list(self.offset, lambda x: printer.print_string(str(x)))
        printer.print_string("] size [")
        printer.print_list(self.size, lambda x: printer.print_string(str(x)))
        printer.print_string("] source offset [")
        printer.print_list(self.source_offset, lambda x: printer.print_string(str(x)))
        printer.print_string(f"] to {list(self.neighbor)}>")

    @classmethod
    def parse_parameters(cls, parser: AttrParser) -> list[Attribute]:
        parser.parse_characters("<")
        parser.parse_characters("at")
        offset = parser.parse_comma_separated_list(
            parser.Delimiter.SQUARE, parser.parse_integer
        )
        parser.parse_characters("size")
        size = parser.parse_comma_separated_list(
            parser.Delimiter.SQUARE, parser.parse_integer
        )
        parser.parse_characters("source")
        parser.parse_characters("offset")
        source_offset = parser.parse_comma_separated_list(
            parser.Delimiter.SQUARE, parser.parse_integer
        )
        parser.parse_characters("to")
        to = parser.parse_comma_separated_list(
            parser.Delimiter.SQUARE, parser.parse_integer
        )
        parser.parse_characters(">")

        return [
            builtin.DenseArrayBase.from_list(builtin.i64, x)
            for x in (offset, size, source_offset, to)
        ]


@irdl_attr_definition
class ShapeAttr(ParametrizedAttribute):
    """
    This represents shape information that is attached to halo operations.

    On the terminology used:

    In each dimension, we are given four points. We abbreviate them in
    annotations to an, bn, cn, dn, with n being the dimension. In 2d, these
    create the following pattern, higher dimensional examples can
    be derived from this:

    a1 b1          c1 d1
    +--+-----------+--+ a0
    |  |           |  |
    +--+-----------+--+ b0
    |  |           |  |
    |  |           |  |
    |  |           |  |
    |  |           |  |
    +--+-----------+--+ c0
    |  |           |  |
    +--+-----------+--+ d0

    We can now name these points:

         - a: buffer_start
         - b: core_start
         - c: core_end
         - d: buffer_end

    This class provides easy getters for these four.

    We can also define some common sizes on this object:

        - buff_size(n) = dn - an
        - core_size(n) = cn - bn
        - halo_size(n, start) = bn - an
        - halo_size(n, end  ) = dn - cn
    """

    name = "dmp.shape_with_halo"

    buff_lb_: ParameterDef[builtin.DenseArrayBase]
    buff_ub_: ParameterDef[builtin.DenseArrayBase]
    core_lb_: ParameterDef[builtin.DenseArrayBase]
    core_ub_: ParameterDef[builtin.DenseArrayBase]

    @property
    def buff_lb(self) -> tuple[int, ...]:
        data = self.buff_lb_.get_values()
        assert isa(data, tuple[int, ...])
        return data

    @property
    def buff_ub(self) -> tuple[int, ...]:
        data = self.buff_ub_.get_values()
        assert isa(data, tuple[int, ...])
        return data

    @property
    def core_lb(self) -> tuple[int, ...]:
        data = self.core_lb_.get_values()
        assert isa(data, tuple[int, ...])
        return data

    @property
    def core_ub(self) -> tuple[int, ...]:
        data = self.core_ub_.get_values()
        assert isa(data, tuple[int, ...])
        return data

    @property
    def dims(self) -> int:
        """
        Number of axis of the data (len(shape))
        """
        return len(self.core_ub)

    @staticmethod
    def from_index_attrs(
        buff_lb: stencil.IndexAttr | Sequence[int],
        core_lb: stencil.IndexAttr | Sequence[int],
        core_ub: stencil.IndexAttr | Sequence[int],
        buff_ub: stencil.IndexAttr | Sequence[int],
    ):
        data_type = builtin.i64
        return ShapeAttr(
            [
                builtin.DenseArrayBase.from_list(data_type, tuple(data))
                for data in (buff_lb, buff_ub, core_lb, core_ub)
            ]
        )

    def buffer_start(self, dim: int) -> int:
        assert dim < self.dims, f"The given DimsHelper only has {self.dims} dimensions"
        return self.buff_lb[dim]

    def core_start(self, dim: int) -> int:
        assert dim < self.dims, f"The given DimsHelper only has {self.dims} dimensions"
        return self.core_lb[dim]

    def buffer_end(self, dim: int) -> int:
        assert dim < self.dims, f"The given DimsHelper only has {self.dims} dimensions"
        return self.buff_ub[dim]

    def core_end(self, dim: int) -> int:
        assert dim < self.dims, f"The given DimsHelper only has {self.dims} dimensions"
        return self.core_ub[dim]

    # Helpers for specific sizes:

    def buff_size(self, dim: int) -> int:
        assert dim < self.dims, f"The given DimsHelper only has {self.dims} dimensions"
        return self.buff_ub[dim] - self.buff_lb[dim]

    def core_size(self, dim: int) -> int:
        assert dim < self.dims, f"The given DimsHelper only has {self.dims} dimensions"
        return self.core_ub[dim] - self.core_lb[dim]

    def halo_size(self, dim: int, at_end: bool = False) -> int:
        assert dim < self.dims, f"The given DimsHelper only has {self.dims} dimensions"
        if at_end:
            return self.buff_ub[dim] - self.core_ub[dim]
        return self.core_lb[dim] - self.buff_lb[dim]

    # parsing / printing

    def print_parameters(self, printer: Printer) -> None:
        dims = zip(self.buff_lb, self.core_lb, self.core_ub, self.buff_ub)
        printer.print_string("<")
        printer.print_string("x".join(f"{list(vals)}" for vals in dims))
        printer.print_string(">")

    @classmethod
    def parse_parameters(cls, parser: AttrParser) -> list[Attribute]:
        """
        Parses the attribute, the format of it is:

        #dmp.shape_with_halo<[a0,b0,c0,d0]x[a1,b1,c1,d1]x...>

        so different from the way it's stored internally.

        This decision was made to improve readability.
        """
        parser.parse_characters("<")
        buff_lb: list[int] = []
        buff_ub: list[int] = []
        core_lb: list[int] = []
        core_ub: list[int] = []

        while True:
            parser.parse_characters("[")
            buff_lb.append(parser.parse_integer())
            parser.parse_characters(",")
            core_lb.append(parser.parse_integer())
            parser.parse_characters(",")
            core_ub.append(parser.parse_integer())
            parser.parse_characters(",")
            buff_ub.append(parser.parse_integer())
            parser.parse_characters("]")
            if parser.parse_optional_characters("x") is None:
                break
        parser.parse_characters(">")

        data_type = builtin.i64
        return [
            builtin.DenseArrayBase.from_list(data_type, data)
            for data in (buff_lb, buff_ub, core_lb, core_ub)
        ]


@irdl_attr_definition
class RankTopoAttr(ParametrizedAttribute):
    """
    This attribute specifies the node layout used to distribute the computation.

    dmp.grid<3x3> means nine ranks organized in a 3x3 grid.

    This allows for higher-dimensional grids as well, e.g. dmp.grid<3x3x3> for
    3-dimensional data.
    """

    name = "dmp.topo"

    shape: ParameterDef[builtin.DenseArrayBase]

    def __init__(self, shape: Sequence[int] | Sequence[builtin.IntAttr]):
        if len(shape) < 1:
            raise ValueError("dmp.grid must have at least one dimension!")
        super().__init__([builtin.DenseArrayBase.from_list(builtin.i64, shape)])

    def as_tuple(self) -> tuple[int, ...]:
        shape = self.shape.get_values()
        assert isa(shape, tuple[int, ...])
        return shape

    def node_count(self) -> int:
        return prod(self.as_tuple())

    @classmethod
    def parse_parameters(cls, parser: AttrParser) -> list[Attribute]:
        parser.parse_characters("<")
        shape: list[int] = [
            parser.parse_integer(allow_negative=False, allow_boolean=False)
        ]

        while parser.parse_optional_punctuation(">") is None:
            parser.parse_shape_delimiter()
            shape.append(
                parser.parse_integer(allow_negative=False, allow_boolean=False)
            )

        return [builtin.DenseArrayBase.from_list(builtin.i64, shape)]

    def print_parameters(self, printer: Printer) -> None:
        printer.print_string("<")
        printer.print_string("x".join(str(x) for x in self.shape.get_values()))
        printer.print_string(">")


class DomainDecompositionStrategy(ParametrizedAttribute, ABC):
    def calc_resize(self, shape: tuple[int, ...]) -> tuple[int, ...]:
        raise NotImplementedError("SlicingStrategy must implement calc_resize!")

    def halo_exchange_defs(self, shape: ShapeAttr) -> Iterable[ExchangeDeclarationAttr]:
        raise NotImplementedError("SlicingStrategy must implement halo_exchange_defs!")

    def comm_layout(self) -> RankTopoAttr:
        raise NotImplementedError("SlicingStrategy must implement comm_count!")


@irdl_attr_definition
class GridSlice2dAttr(DomainDecompositionStrategy):
    """
    Takes a grid with two or more dimensions, slices it along the first two into equally
    sized segments.
    """

    name = "dmp.grid_slice_2d"

    topology: ParameterDef[RankTopoAttr]

    diagonals: ParameterDef[builtin.BoolAttr]

    def __init__(self, topo: tuple[int, ...]):
        super().__init__(
            [RankTopoAttr(topo), builtin.BoolAttr.from_int_and_width(0, 1)]
        )

    def _verify(self):
        assert len(self.topology.as_tuple()) >= 2, (
            "GridSlice2d requires at least two dimensions"
        )

    def calc_resize(self, shape: tuple[int, ...]) -> tuple[int, ...]:
        assert len(shape) >= 2, "GridSlice2d requires at least two dimensions"
        for size, node_count in zip(shape, self.topology.as_tuple()):
            assert size % node_count == 0, (
                "GridSlice2d requires domain be neatly divisible by shape"
            )
        return (
            *(
                size // node_count
                for size, node_count in zip(shape, self.topology.as_tuple())
            ),
            *(size for size in shape[2:]),
        )

    def halo_exchange_defs(self, shape: ShapeAttr) -> Iterable[ExchangeDeclarationAttr]:
        yield from _flat_face_exchanges_for_dim(shape, 0)

        yield from _flat_face_exchanges_for_dim(shape, 1)

        if self.diagonals.value.data:
            raise NotImplementedError("Diagonals support not implemented yet")

    def comm_layout(self) -> RankTopoAttr:
        return RankTopoAttr(self.topology.as_tuple())


@irdl_attr_definition
class GridSlice3dAttr(DomainDecompositionStrategy):
    """
    Takes a grid with two or more dimensions, slices it along the first three.
    """

    name = "dmp.grid_slice_3d"

    topology: ParameterDef[RankTopoAttr]

    diagonals: ParameterDef[builtin.BoolAttr]

    def __init__(self, topo: tuple[int, ...]):
        super().__init__(
            [RankTopoAttr(topo), builtin.BoolAttr.from_int_and_width(0, 1)]
        )

    def _verify(self):
        assert len(self.topology.as_tuple()) >= 3, (
            "GridSlice3d requires at least three dimensions"
        )

    def calc_resize(self, shape: tuple[int, ...]) -> tuple[int, ...]:
        assert len(shape) >= 3, "GridSlice3d requires at least two dimensions"
        for size, node_count in zip(shape, self.topology.as_tuple()):
            assert size % node_count == 0, (
                "GridSlice3d requires domain be neatly divisible by shape"
            )
        return (
            *(
                size // node_count
                for size, node_count in zip(shape, self.topology.as_tuple())
            ),
            *(size for size in shape[3:]),
        )

    def halo_exchange_defs(self, shape: ShapeAttr) -> Iterable[ExchangeDeclarationAttr]:
        yield from _flat_face_exchanges_for_dim(shape, 0)

        yield from _flat_face_exchanges_for_dim(shape, 1)

        yield from _flat_face_exchanges_for_dim(shape, 2)

        if self.diagonals.value.data:
            raise NotImplementedError("Diagonals support not implemented yet")

    def comm_layout(self) -> RankTopoAttr:
        return RankTopoAttr(self.topology.as_tuple())


def _flat_face_exchanges_for_dim(
    shape: ShapeAttr, axis: int
) -> tuple[ExchangeDeclarationAttr, ...]:
    """
    Generate the two exchange delcarations to exchange the faces on the
    axis "axis".
    """
    dimensions = shape.dims
    assert axis <= dimensions

    def coords(where: Literal["start", "end"]) -> Iterable[tuple[tuple[int, int], ...]]:
        """
        Generate a series of swaps that need to be performed to exchange along "axis".

        A swap is a set of (lb,ub) tuples, one per axis of shape.

        Takes either "start" or "end" to signify if the lower (buffer start to core start) or upper
        (core end to buffer end) parts of the halo should be exchanged.

        We need to make sure that if core_size is smaller than halo size, we emit multiple exchanges.

        We need to make sure that we emit the exchanges in a way that the closest neighbor is emitted first.
        """
        # we may need to issue multiple swaps per direction, if the core size is smaller than the
        # exchanged size. This is tracked in the "slice" variable.
        slice = 0

        while True:
            swap: list[tuple[int, int]] = []
            for d in range(dimensions):
                # for the dim we want to exchange, return exchanges need to exchange either start or end
                # halo regions
                if d == axis:
                    core_size = shape.core_size(d)
                    if where == "start":
                        # where == "start" halo goes from buffer start to core start
                        # the window of data we want to send starts here
                        start = shape.buffer_start(d)
                        # calculate where the current slice starts (lowest index, no lower than start)
                        slice_start = max(
                            start, shape.core_start(d) - (core_size * (slice + 1))
                        )
                        # calculate where the current slice ends (highest index, no higher than core_start)
                        # because slice >= 0
                        slice_end = max(
                            start, shape.core_start(d) - (core_size * slice)
                        )

                        # stop swapping if swap is empty
                        if slice_end == slice_start:
                            return
                        swap.append((slice_start, slice_end))
                    else:
                        # where == "end" halo goes from core end to buffer end

                        # the window of data we want to send ends here (highest index)
                        end = shape.buffer_end(d)
                        # calculate where the current slice starts (lowest index, no lower than start)
                        # because slice >= 0, and no higher than end
                        slice_start = min(end, shape.core_end(d) + (core_size * slice))
                        # calculate where the current slice ends (highest index, no higher than core_start)
                        slice_end = min(
                            end, shape.core_end(d) + (core_size * (slice + 1))
                        )

                        # stop swapping if swap is empty
                        if slice_end == slice_start:
                            return
                        swap.append((slice_start, slice_end))

                else:
                    # for the sliced regions, "extrude" from core
                    # this way we don't exchange edges
                    swap.append((shape.core_start(d), shape.core_end(d)))

            slice += 1
            yield tuple(swap)

    return (
        # towards positive dim:
        *(
            ExchangeDeclarationAttr.from_points(
                ex1_coords,
                axis,
                dir_sign=1,
                neighbor_offset=i + 1,
            )
            for i, ex1_coords in enumerate(coords("end"))
        ),
        # towards negative dim:
        *(
            ExchangeDeclarationAttr.from_points(
                ex2_coords,
                axis,
                dir_sign=-1,
                neighbor_offset=i + 1,
            )
            for i, ex2_coords in enumerate(coords("start"))
        ),
    )


class SwapOpHasShapeInferencePatterns(HasShapeInferencePatternsTrait):
    @classmethod
    def get_shape_inference_patterns(cls):
        from xdsl.transforms.shape_inference_patterns.dmp import (
            DmpSwapShapeInference,
            DmpSwapSwapsInference,
        )

        return (DmpSwapShapeInference(), DmpSwapSwapsInference())


class SwapOpMemoryEffect(MemoryEffect):
    """
    Side effect implementation of dmp.swap.
    """

    @classmethod
    def get_effects(cls, op: Operation) -> set[EffectInstance]:
        op = cast(SwapOp, op)
        # If it's operating in value-semantic mode, it has no side effects.
        if op.swapped_values:
            return set()
        # If it's operating in reference-semantic mode, it reads and writes to its field.
        # TODO: consider the empty swaps case at some point.
        # Right now, it relies on it before inferring them, so not very safe.
        # But it could be an elegant way to generically simplify those.
        return {
            EffectInstance(MemoryEffectKind.WRITE, op.input_stencil),
            EffectInstance(MemoryEffectKind.READ, op.input_stencil),
        }


@irdl_op_definition
class SwapOp(IRDLOperation):
    """
    Declarative swap of memref regions.
    """

    name = "dmp.swap"

    input_stencil = operand_def(stencil.StencilTypeConstr)
    swapped_values = opt_result_def(stencil.TempType[Attribute])

    swaps = attr_def(builtin.ArrayAttr[ExchangeDeclarationAttr])

    strategy = attr_def(DomainDecompositionStrategy)

    traits = traits_def(SwapOpHasShapeInferencePatterns(), SwapOpMemoryEffect())

    def verify_(self) -> None:
        if self.swapped_values:
            if isinstance(self.input_stencil.type, stencil.FieldType):
                raise VerifyException(
                    "dmp.swap_op cannot have a result if input is a field"
                )
        else:
            if isinstance(self.input_stencil.type, stencil.TempType):
                raise VerifyException(
                    "dmp.swap_op must have a result if input is a temporary"
                )

    @staticmethod
    def get(
        input_stencil: SSAValue | Operation,
        strategy: DomainDecompositionStrategy,
        swaps: builtin.ArrayAttr[ExchangeDeclarationAttr] | None = None,
    ):
        input_type = SSAValue.get(input_stencil).type

        result_types = (
            input_type if isa(input_type, stencil.TempType[Attribute]) else None
        )

        if swaps is None:
            swaps = builtin.ArrayAttr[ExchangeDeclarationAttr](())

        return SwapOp.build(
            operands=[input_stencil],
            result_types=[result_types],
            attributes={
                "strategy": strategy,
                "swaps": swaps,
            },
        )


DMP = Dialect(
    "dmp",
    [
        SwapOp,
    ],
    [
        ExchangeDeclarationAttr,
        ShapeAttr,
        RankTopoAttr,
        GridSlice2dAttr,
        GridSlice3dAttr,
    ],
)



import ast
from typing import Any


def block(*params: Any):
    """
    Decorator used to mark function as a basic block.
    ```
    def foo(a: int) -> int:
        @block
        def bb0(x: int):
            y: int = x + 2
            bb1(y)

        @block
        def bb1(z: int):
            return z

        # Entry-point.
        bb0(a)
    ```
    """

    def decorate(*params: Any):
        return None

    return decorate


def is_block(node: ast.FunctionDef) -> bool:
    return (
        len(node.decorator_list) == 1
        and isinstance(name := node.decorator_list[0], ast.Name)
        and name.id == "block"
    )



"""
CIRCT’s seq dialect.

See external [documentation](https://circt.llvm.org/docs/Dialects/Seq/).
"""

from enum import Enum
from typing import ClassVar

from xdsl.dialects.builtin import (
    IntegerAttr,
    IntegerType,
    TypeAttribute,
    i1,
)
from xdsl.dialects.hw import InnerSymAttr
from xdsl.ir import Data, Dialect, Operation, SSAValue
from xdsl.irdl import (
    AnyAttr,
    AttrSizedOperandSegments,
    IRDLOperation,
    ParametrizedAttribute,
    VarConstraint,
    attr_def,
    irdl_attr_definition,
    irdl_op_definition,
    operand_def,
    opt_attr_def,
    opt_operand_def,
    result_def,
)
from xdsl.parser import AttrParser, Parser
from xdsl.printer import Printer
from xdsl.utils.exceptions import VerifyException


@irdl_attr_definition
class ClockType(ParametrizedAttribute, TypeAttribute):
    """
    A type for clock-carrying wires. Signals which can be used to drive the clock input of sequential operations.
    """

    name = "seq.clock"


clock = ClockType()


@irdl_op_definition
class ClockDividerOp(IRDLOperation):
    """Produces a clock divided by a power of two"""

    name = "seq.clock_div"

    pow2 = attr_def(IntegerAttr)
    clockIn = operand_def(ClockType)
    clockOut = result_def(ClockType)

    def __init__(self, clockIn: SSAValue | Operation, pow2: int | IntegerAttr):
        if isinstance(pow2, int):
            pow2 = IntegerAttr(pow2, IntegerType(8))
        super().__init__(
            operands=[clockIn], attributes={"pow2": pow2}, result_types=[clock]
        )

    def verify_(self) -> None:
        if self.pow2.type != IntegerType(8):
            raise VerifyException("pow2 has to be an 8-bit signless integer")
        if self.pow2.value.data.bit_count() != 1:
            raise VerifyException(
                f"divider value {self.pow2.value.data} is not a power of 2"
            )

    @classmethod
    def parse(cls, parser: Parser):
        input_ = parser.parse_operand()
        parser.parse_keyword("by")
        divider = parser.parse_integer()
        return cls(input_, divider)

    def print(self, printer: Printer):
        printer.print(" ")
        printer.print_operand(self.clockIn)
        printer.print(" by ")
        printer.print(self.pow2.value.data)


@irdl_op_definition
class CompRegOp(IRDLOperation):
    """
    Register a value, storing it for one cycle.
    """

    name = "seq.compreg"

    DATA_TYPE: ClassVar = VarConstraint("DataType", AnyAttr())

    inner_sym = opt_attr_def(InnerSymAttr)
    input = operand_def(DATA_TYPE)
    clk = operand_def(clock)
    reset = opt_operand_def(i1)
    reset_value = opt_operand_def(DATA_TYPE)
    power_on_value = opt_operand_def(DATA_TYPE)
    data = result_def(DATA_TYPE)

    irdl_options = [AttrSizedOperandSegments()]

    assembly_format = (
        "(`sym` $inner_sym^)? $input `,` $clk "
        "(`reset` $reset^ `,` $reset_value)? "
        "(`powerOn` $power_on_value^)? "
        "attr-dict `:` type($input)"
    )

    def __init__(
        self,
        input: SSAValue,
        clk: SSAValue,
        reset: tuple[SSAValue, SSAValue] | None = None,
        power_on_value: SSAValue | None = None,
    ):
        super().__init__(
            operands=[
                input,
                clk,
                reset[0] if reset is not None else None,
                reset[1] if reset is not None else None,
                power_on_value,
            ],
            result_types=[input.type],
        )

    def verify_(self):
        if (self.reset is None) != (self.reset_value is None):
            raise VerifyException("Both reset and reset_value must be set when one is")


class ClockConstAttrData(Enum):
    LOW = 0
    HIGH = 1


@irdl_attr_definition
class ClockConstAttr(Data[ClockConstAttrData]):
    """
    Clock constant.

    This attribute diverges slightly from the upstream implementation
    as xDSL does not allow completely unstructured parsing and printing
    of attributes (for good reasons).
    """

    name = "seq.clock_constant"

    @classmethod
    def parse_parameter(cls, parser: AttrParser) -> ClockConstAttrData:
        with parser.in_angle_brackets():
            return ClockConstAttr.parse_parameter_free_standing(parser)

    @classmethod
    def parse_parameter_free_standing(cls, parser: AttrParser) -> ClockConstAttrData:
        if parser.parse_optional_keyword("low") is not None:
            return ClockConstAttrData.LOW
        if parser.parse_optional_keyword("high") is not None:
            return ClockConstAttrData.HIGH
        parser.raise_error("Expected either low or high clock value")

    def print_parameter(self, printer: Printer) -> None:
        with printer.in_angle_brackets():
            self.print_parameter_free_standing(printer)

    def print_parameter_free_standing(self, printer: Printer) -> None:
        match self.data:
            case ClockConstAttrData.LOW:
                printer.print("low")
            case ClockConstAttrData.HIGH:
                printer.print("high")


@irdl_op_definition
class ConstClockOp(IRDLOperation):
    """
    The constant operation produces a constant clock value.
    """

    name = "seq.const_clock"

    value = attr_def(ClockConstAttr)
    result = result_def(clock)

    @classmethod
    def parse(cls, parser: Parser) -> "ConstClockOp":
        value = ClockConstAttr(ClockConstAttr.parse_parameter_free_standing(parser))
        attrs = parser.parse_optional_attr_dict_with_reserved_attr_names(("value",))
        attrs_data = dict(attrs.data) if attrs is not None else {}
        attrs_data["value"] = value
        return ConstClockOp.create(attributes=attrs_data, result_types=[clock])

    def print(self, printer: Printer):
        printer.print(" ")
        self.value.print_parameter_free_standing(printer)
        printer.print_op_attributes(self.attributes, reserved_attr_names=("value",))


Seq = Dialect(
    "seq",
    [
        ClockDividerOp,
        CompRegOp,
        ConstClockOp,
    ],
    [
        ClockType,
        ClockConstAttr,
    ],
)



import marimo

__generated_with = "0.11.10"
app = marimo.App()


@app.cell(hide_code=True)
def _():
    import marimo as mo
    from xdsl.builder import Builder, InsertPoint
    from xdsl.utils import marimo as xmo
    from xdsl.dialects.arith import AddiOp
    from xdsl.dialects.builtin import ModuleOp, IntegerType, IntegerAttr
    from xdsl.dialects.arith import ConstantOp
    return (
        AddiOp,
        Builder,
        ConstantOp,
        InsertPoint,
        IntegerAttr,
        IntegerType,
        ModuleOp,
        mo,
        xmo,
    )


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""# Builders""")
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""Builders are used to insert new operations in an existing block. While they also exist in MLIR, they work a bit differently in xDSL. In MLIR, builders are used to create operations, while in xDSL builders are used to insert already created operations.""")
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""## Operation Constructors""")
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""
        In xDSL, ops often define a custom `__init__` constructor to make operation creation easier.

        For example, here is the API for creating an `arith.constant` instance:
        """
    )
    return


@app.cell
def _(ConstantOp, IntegerAttr):
    c0 = ConstantOp(IntegerAttr(0, 64))
    c0
    return (c0,)


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""Similarly, here is the API for creating an `arith.addi` instance:""")
    return


@app.cell
def _(AddiOp, c0):
    addi = AddiOp(c0.result, c0.result)
    addi
    return (addi,)


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""Generally, it is recommended to look at the definition of the operation you would like to create to see the API. Here is the xDSL generated documentation for the [`arith` dialect](https://xdsl.readthedocs.io/en/latest/reference/xdsl.dialects.arith.html) and the [`scf` dialect](https://xdsl.readthedocs.io/en/latest/reference/xdsl.dialects.scf.html).""")
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""## Using the `Builder`""")
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""
        A `Builder` insert new operations at a given insertion point (A position before an operation, or at the end of a block). Whenever a builder inserts a new operation, it updates its insertion point to be after the inserted operation.

        The `Builder` constructor takes a single argument, which is an `InsertPoint`. An `InsertPoint` can be created using the static methods `before`, `after`, `at_start`, and `at_end`.

        Let's look at the following program:
        """
    )
    return


@app.cell(hide_code=True)
def _(ConstantOp, IntegerAttr, IntegerType, ModuleOp, xmo):
    module = ModuleOp([
        ConstantOp(IntegerAttr(0, IntegerType(64))),
        ConstantOp(IntegerAttr(1, IntegerType(64))),
        ConstantOp(IntegerAttr(2, IntegerType(64))),
    ])

    xmo.module_html(module)
    return (module,)


@app.cell(hide_code=True)
def _(mo):
    mo.md("""The following code inserts the constant `3` and `4` in the IR using the builders.""")
    return


@app.cell
def _(Builder, ConstantOp, InsertPoint, IntegerAttr, IntegerType, module):
    # Clone the module to only do modifications locally.
    # Otherwise this messes up other cells using `module`.
    module_cloned = module.clone()
    block_cloned = module_cloned.body.block
    cst0_cloned = module_cloned.body.ops.first
    cst1_cloned = cst0_cloned.next_op

    # Insert 3 between 0 and 1
    builder1 = Builder(InsertPoint.before(cst1_cloned))
    builder1.insert(ConstantOp(IntegerAttr(3, IntegerType(64))))

    # Insert 4 and 5 at the end of the block
    builder2 = Builder(InsertPoint.at_end(block_cloned))
    builder2.insert(ConstantOp(IntegerAttr(4, IntegerType(64))))
    builder2.insert(ConstantOp(IntegerAttr(5, IntegerType(64))))

    None
    return (
        block_cloned,
        builder1,
        builder2,
        cst0_cloned,
        cst1_cloned,
        module_cloned,
    )


@app.cell
def _(module_cloned):
    module_cloned
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""
        ## Small exercise

        As a simple task, insert the program so the operations are read in order `0, 1, 2, 3, 4, 5, 6, 7`
        """
    )
    return


@app.cell(hide_code=True)
def _(ConstantOp, IntegerAttr, IntegerType, ModuleOp, xmo):
    module2 = ModuleOp([
        ConstantOp(IntegerAttr(1, IntegerType(64))),
        ConstantOp(IntegerAttr(4, IntegerType(64))),
        ConstantOp(IntegerAttr(6, IntegerType(64))),
    ])

    xmo.module_html(module2)
    return (module2,)


@app.cell
def _(module2):
    module2_cloned = module2.clone()

    block = module2_cloned.body.block
    cst1 = block.first_op
    cst4 = cst1.next_op
    cst6 = cst4.next_op

    # Use the builder here
    return block, cst1, cst4, cst6, module2_cloned


if __name__ == "__main__":
    app.run()



from xdsl.utils.hasher import Hasher


def test_hasher():
    h = Hasher()
    h.combine(1)
    hash1 = h.hash
    h.combine(2)

    j = Hasher(seed=hash1)

    assert h.hash != j.hash

    j.combine(2)

    assert h.hash == j.hash



# RUN: python %s | filecheck %s

from xdsl.frontend.pyast.context import CodeContext
from xdsl.frontend.pyast.dialects.builtin import i1, i32
from xdsl.frontend.pyast.exception import CodeGenerationException
from xdsl.frontend.pyast.program import FrontendProgram

p = FrontendProgram()
with CodeContext(p):
    # CHECK: cf.assert %{{.*}}, ""
    def test_assert_I(cond: i1):
        assert cond
        return

    # CHECK: cf.assert %{{.*}}, "some message"
    def test_assert_II(cond: i1):
        assert cond, "some message"
        return


p.compile(desymref=False)
print(p.textual_format())

try:
    with CodeContext(p):
        # CHECK: Expected a string constant for assertion message, found 'ast.Name'
        def test_assert_message_type(cond: i1, a: i32):
            assert cond, a
            return

    p.compile(desymref=False)
    print(p.textual_format())
except CodeGenerationException as e:
    print(e.msg)



import pytest

from xdsl.utils.exceptions import PassPipelineParseError
from xdsl.utils.parse_pipeline import (
    PipelinePassSpec,
    parse_pipeline,
)


def test_pass_parser():
    passes = list(
        parse_pipeline(
            'mlir-opt[cse],pass-1,pass-2{arg1=1 arg2=test,test2,3 arg3="test-str,2,3" '
            "arg-4=-34.4e-12 no-val-arg},mlir-opt[cse],pass-3{thing=2d-grid},mlir-opt[cse]"
        )
    )

    assert passes == [
        PipelinePassSpec(
            "mlir-opt",
            {
                "arguments": (
                    "--mlir-print-op-generic",
                    "--allow-unregistered-dialect",
                    "-p",
                    "builtin.module(cse)",
                )
            },
        ),
        PipelinePassSpec("pass-1", {}),
        PipelinePassSpec(
            "pass-2",
            {
                "arg1": (1,),
                "arg2": ("test", "test2", 3),
                "arg3": ("test-str,2,3",),
                "arg-4": (-3.44e-11,),
                "no-val-arg": (),
            },
        ),
        PipelinePassSpec(
            "mlir-opt",
            {
                "arguments": (
                    "--mlir-print-op-generic",
                    "--allow-unregistered-dialect",
                    "-p",
                    "builtin.module(cse)",
                )
            },
        ),
        PipelinePassSpec(
            "pass-3",
            {
                "thing": ("2d-grid",),
            },
        ),
        PipelinePassSpec(
            "mlir-opt",
            {
                "arguments": (
                    "--mlir-print-op-generic",
                    "--allow-unregistered-dialect",
                    "-p",
                    "builtin.module(cse)",
                )
            },
        ),
    ]


@pytest.mark.parametrize(
    "spec",
    [
        PipelinePassSpec("empty", {}),
        PipelinePassSpec(
            "pass-2",
            {
                "arg1": (1,),
                "arg2": (
                    "test",
                    "test2",
                    3,
                ),
                "arg3": ("test-str,2,3",),
                "arg-4": (-3.44e-11,),
                "no-val-arg": (),
            },
        ),
        PipelinePassSpec(
            "pass-3",
            {
                "thing": ("2d-grid",),
            },
        ),
    ],
)
def test_spec_printer(spec: PipelinePassSpec):
    text = str(spec)
    passes = list(parse_pipeline(text))
    assert len(passes) == 1
    assert passes[0] == spec


def test_invalid_mlir_pipeline():
    with pytest.raises(
        PassPipelineParseError,
        match="Expected `mlir-opt` to mark an MLIR pipeline here",
    ):
        list(parse_pipeline("canonicalize[cse]"))



#!/usr/bin/env python3
"""Microbenchmark properties of the xDSL implementation."""

import importlib

import xdsl.dialects.arith
import xdsl.dialects.builtin


class LoadDialects:
    """Benchmark loading dialects in xDSL."""

    def time_arith_load(self) -> None:
        """Time loading the `arith` dialect.

        Note that this must be done with `importlib.reload` rather than just
        directly importing with `from xdsl.dialects.arith import Arith` to avoid
        tests interacting with each other.
        """
        importlib.reload(xdsl.dialects.arith)

    def time_builtin_load(self) -> None:
        """Time loading the `builtin` dialect."""
        importlib.reload(xdsl.dialects.builtin)


if __name__ == "__main__":
    from bench_utils import Benchmark, profile

    LOAD_DIALECTS = LoadDialects()

    profile(
        {
            "LoadDialects.arith_load": Benchmark(LOAD_DIALECTS.time_arith_load),
            "LoadDialects.builtin_load": Benchmark(LOAD_DIALECTS.time_builtin_load),
        }
    )



import marimo

__generated_with = "0.11.19"
app = marimo.App(width="medium")


@app.cell(hide_code=True)
def _():
    import marimo as mo

    # xDSL should be available in the environment
    from xdsl.dialects.arith import Arith
    from xdsl.dialects.builtin import Builtin
    from xdsl.context import Context
    from xdsl.printer import Printer

    from xdsl.dialects.builtin import i32
    from xdsl.utils.exceptions import VerifyException

    # Context, containing information about the registered dialects
    context = Context()

    # Some useful dialects
    context.load_dialect(Arith)
    context.load_dialect(Builtin)

    # Printer used to pretty-print MLIR data structures
    printer = Printer()
    return (
        Arith,
        Builtin,
        Context,
        Printer,
        VerifyException,
        context,
        i32,
        mo,
        printer,
    )


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""
        # IRDL tutorial

        ## An Intermediate Representation Definition Language (IRDL) for SSA Compilers
        """
    )
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""
        ## Introduction

        xDSL is an extensible compiler, meaning that new operations, attributes, and types can be added. xDSL provides an embedded DSL, IRDL, to define new dialects.
        This tutorial aims to show the different features IRDL has, and presents examples on how to use them.
        """
    )
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""
        ### Attribute constraints

        Attribute constraints represent invariants over attributes, and are an important concept for defining new attributes and operations. In practice, an attribute constraint is a child class of `AttrConstraint` that implements a `verify` method. The method takes an attribute to verify as parameter, and a dictionary associating constraint variables to attributes. `verify` does not return anything, but raises an exception if the invariant is not respected.
        """
    )
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""
        #### Any Constraint

        An `Any` constraint will never trigger an exception, and will always pass:
        """
    )
    return


@app.cell
def _():
    from xdsl.dialects.builtin import IndexType, IntegerType, StringAttr, i64
    from xdsl.irdl import AnyAttr, ConstraintContext

    # Construct the constraint
    any_constraint = AnyAttr()

    # This will pass without triggering an exception
    any_constraint.verify(i64, ConstraintContext())
    any_constraint.verify(StringAttr("ga"), ConstraintContext())
    return (
        AnyAttr,
        ConstraintContext,
        IndexType,
        IntegerType,
        StringAttr,
        any_constraint,
        i64,
    )


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""
        #### Equality Constraint

        An equality constraint ensures that the attribute is equal to one provided to the constraint:
        """
    )
    return


@app.cell
def _(ConstraintContext, i32, i64):
    from xdsl.irdl import EqAttrConstraint

    # Construct the constraint
    eq_constraint = EqAttrConstraint(i64)

    # This will pass without triggering an exception
    eq_constraint.verify(i64, ConstraintContext())

    # This will trigger an exception
    try:
        eq_constraint.verify(i32, ConstraintContext())
    except Exception as e:
        print(e)
    return EqAttrConstraint, eq_constraint


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""
        #### Base Attribute Constraint

        A base attribute constraint ensures that the attribute base type is equal to an expected attribute base type:
        """
    )
    return


@app.cell
def _(ConstraintContext, StringAttr, VerifyException, i32):
    from xdsl.dialects.builtin import IntAttr
    from xdsl.irdl import BaseAttr

    # Construct the constraint
    base_constraint = BaseAttr(StringAttr)

    # This will pass without triggering an exception
    base_constraint.verify(StringAttr("ga"), ConstraintContext())
    base_constraint.verify(StringAttr("bu"), ConstraintContext())

    # This will trigger an exception
    try:
        base_constraint.verify(i32, ConstraintContext())
    except VerifyException as e:
        print(e)
    return BaseAttr, IntAttr, base_constraint


@app.cell
def _(ConstraintContext, IntAttr, VerifyException, base_constraint):
    # This too
    try:
        base_constraint.verify(IntAttr(3), ConstraintContext())
    except VerifyException as e:
        print(e)
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""
        #### Attribute Constraint Coercion

        To simplify the definitions of constraints, constraint constructors expecting an attribute constraints will coerce `Attribute` to an equality attribute constraint, and will coerce an `Attribute` type to a base attribute constraint. this is done using the `attr_constr_coercion` function:
        """
    )
    return


@app.cell
def _(BaseAttr, EqAttrConstraint, StringAttr, i32):
    from xdsl.irdl import AnyOf, attr_constr_coercion

    assert attr_constr_coercion(i32) == EqAttrConstraint(i32)
    assert attr_constr_coercion(StringAttr) == BaseAttr(StringAttr)
    assert AnyOf([i32, StringAttr]) == AnyOf([EqAttrConstraint(i32), BaseAttr(StringAttr)])
    return AnyOf, attr_constr_coercion


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""
        #### Or Constraint

        An `or` constraint ensures that one of the given attribute constraints is satisfied by the attribute:
        """
    )
    return


@app.cell
def _(AnyOf, StringAttr, i32, i64):
    # Construct the constraint. Note that we are using the coercion defined previously.
    or_constraint = AnyOf([i32, StringAttr])

    # This will pass without triggering an exception, since the first constraint is satisfied
    or_constraint.verify(i32)

    # This will pass without triggering an exception, since the second constraint is satisfied
    or_constraint.verify(StringAttr("ga"))
    or_constraint.verify(StringAttr("bu"))

    # This will trigger an exception, since none of the constraints are satisfied
    try:
        or_constraint.verify(i64)
    except Exception as e:
        print(e)
    return (or_constraint,)


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""
        #### Parametric Attribute Constraint

        A parametric attribute constraint is satisfied by parametric attributes of a certain base type. In addition, parametric attribute constraints specify constraints for each of the parameters of the attribute:
        """
    )
    return


@app.cell
def _(ConstraintContext, IntAttr, VerifyException, i32, i64):
    from xdsl.dialects.builtin import IntegerAttr
    from xdsl.irdl import ParamAttrConstraint

    # Construct the constraint. Note that we are using the coercion defined previously.
    param_constraint = ParamAttrConstraint(IntegerAttr, [IntAttr, i32])

    # This will pass without triggering an exception.
    param_constraint.verify(IntegerAttr(IntAttr(42), i32), ConstraintContext())
    param_constraint.verify(IntegerAttr(IntAttr(23), i32), ConstraintContext())

    # This will trigger an exception since the attribute type is not the expected one
    try:
        param_constraint.verify(i64, ConstraintContext())
    except VerifyException as e:
        print(e)
    return IntegerAttr, ParamAttrConstraint, param_constraint


@app.cell
def _(
    ConstraintContext,
    IntAttr,
    IntegerAttr,
    VerifyException,
    i64,
    param_constraint,
):
    # This will trigger an exception since the second parameter constraint is not satisfied

    try:
        param_constraint.verify(IntegerAttr(IntAttr(42), i64), ConstraintContext())
    except VerifyException as e:
        print(e)
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""
        #### Constraint Variables

        Constraint variables are used to specify equality between attributes in operation and attribute definitions. They also contain a constraint that must be satisfied.
        The first time a constraint variable is used, it will check that the constraint is satisfied. If it is satisfied, it sets the variable to the given attribute. If a constraint variable is already set, it will check that the given attribute is equal to the one already set. Two constraint variables with the same name are considered equal, and are expected to carry the same constraint.
        """
    )
    return


@app.cell
def _(BaseAttr, ConstraintContext, IndexType, IntegerType, VerifyException):
    from xdsl.irdl import VarConstraint

    var_constraint = VarConstraint("T", BaseAttr(IntegerType))
    constraint_context = ConstraintContext()

    # The underlying constraint is not satisfied by the given attribute
    try:
        var_constraint.verify(IndexType(), constraint_context)
    except VerifyException as e:
        print(e)
    return VarConstraint, constraint_context, var_constraint


@app.cell
def _(constraint_context, i32, var_constraint):
    # The constraint sets the constraint variable if the attribute verifies
    var_constraint.verify(i32, constraint_context)
    print(constraint_context.get_variable("T"))
    return


@app.cell
def _(VerifyException, constraint_context, i64, var_constraint):
    # Since the variable is set, the constraint can now only be satisfied by the same attribute
    try:
        var_constraint.verify(i64, constraint_context)
    except VerifyException as e:
        print(e)
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""
        #### Custom Constraints

        Users can define their own constraints for their own types. For instance, here is the definition of the `ArrayOfConstraint` constraint, which expects an `ArrayAttr` where all elements of the array satisfy a specific constraint:
        """
    )
    return


@app.cell
def _(ConstraintContext, IntAttr, VerifyException, attr_constr_coercion):
    from dataclasses import dataclass

    from xdsl.dialects.builtin import ArrayAttr
    from xdsl.ir import Attribute
    from xdsl.irdl import AttrConstraint
    from xdsl.utils.hints import isa


    @dataclass(frozen=True)
    class ArrayOfConstraint(AttrConstraint):
        # The constraint that needs to be satisfied by all elements of the array
        elem_constr: AttrConstraint

        # The custom init applies the attribute constraint coercion
        def __init__(self, constr: Attribute | type[Attribute] | AttrConstraint):
            object.__setattr__(self, "elem_constr", attr_constr_coercion(constr))

        # Check that an attribute satisfies the constraints
        def verify(self, attr: Attribute, constraint_context: ConstraintContext) -> None:
            # We first check that the attribute is an ArrayAttr
            if not isa(attr, ArrayAttr[Attribute]):
                raise VerifyException(f"expected attribute ArrayData but got {attr}")

            # We check the constraint for all elements in the array
            for e in attr.data:
                self.elem_constr.verify(e, constraint_context)


    array_constraint = ArrayOfConstraint(IntAttr)

    # This will pass without triggering an exception
    array_constraint.verify(ArrayAttr([IntAttr(42)]), ConstraintContext())
    array_constraint.verify(ArrayAttr([IntAttr(3), IntAttr(7)]), ConstraintContext())
    return (
        ArrayAttr,
        ArrayOfConstraint,
        AttrConstraint,
        Attribute,
        array_constraint,
        dataclass,
        isa,
    )


@app.cell
def _(ConstraintContext, array_constraint, i32):
    # This will trigger an exception, since the attribute is not an array
    try:
        array_constraint.verify(i32, ConstraintContext())
    except Exception as e:
        print(e)
    return


@app.cell
def _(ArrayAttr, ConstraintContext, IntAttr, StringAttr, array_constraint):
    # This will trigger an exception, since the array contains attribute that do not satisfies the constraint
    try:
        array_constraint.verify(
            ArrayAttr([IntAttr(42), StringAttr("ga")]), ConstraintContext()
        )
    except Exception as e:
        print(e)
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""### Attribute Definition""")
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""
        #### Data Attributes

        `Data` attribute types are defined by inheriting the python `Data` class. Each data attribute definition should define a `name`, and two methods for conversion to a string representation. Here is for example the definition of `IntAttr`:
        """
    )
    return


@app.cell
def _(Printer, printer):
    from xdsl.ir import Data
    from xdsl.irdl import irdl_attr_definition
    from xdsl.parser import AttrParser


    @irdl_attr_definition
    class MyIntAttr(Data[int]):
        name = "my_int"

        @classmethod
        def parse_parameter(cls, parser: AttrParser) -> int:
            data = parser.parse_integer()
            return data

        def print_parameter(self, printer: Printer) -> None:
            printer.print_string(f"{self.data}")


    MyIntAttr(3).print_parameter(printer)
    return AttrParser, Data, MyIntAttr, irdl_attr_definition


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""
        #### Parametrized Attributes

        `ParametrizedAttribute` attribute types are defined using the `irdl_attr_definition` decorator on a class. Such class should contain a `name` field specifying the attribute name. Parameters are added to attribute definitions by defining fields containing a `ParameterDef`. The field names correspond to the parameter names, and `ParameterDef` contains a constraint that should be respected by this parameter. The order of the fields correspond to the order of the parameters when using the attribute. Upon construction of an attribute, all constraints will be checked, and an exception will be raised if the invariants are not satisfied.

        Here is an example of an integer type definition:
        """
    )
    return


@app.cell
def _(IntAttr, StringAttr, irdl_attr_definition):
    from xdsl.ir import ParametrizedAttribute
    from xdsl.irdl import ParameterDef


    # Represent an integer type with a given bitwidth
    @irdl_attr_definition
    class MyIntegerType(ParametrizedAttribute):
        # Name of the type. This is used for printing and parsing.
        name = "integer_type"

        # Only parameter of the type, with an `EqAttrConstraint` constraint.
        # Note the use of the attribute constraint coercion.
        width: ParameterDef[IntAttr]


    my_i32 = MyIntegerType([IntAttr(32)])

    # This will trigger an exception, since the attribute only expect a single attribute
    try:
        MyIntegerType([IntAttr(32), IntAttr(64)])
    except Exception as e:
        print(e)

    # This will trigger an exception, since the attribute is not an IntAttr
    try:
        MyIntegerType([StringAttr("ga")])
    except Exception as e:
        print(e)
    return MyIntegerType, ParameterDef, ParametrizedAttribute, my_i32


@app.cell
def _(mo):
    mo.md(r"""Each parameter can be accessed using the `parameters` field.""")
    return


@app.cell
def _(my_i32):
    my_i32.parameters
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""Our attribute definition also defines accessors for each parameter based on the name given in the `ParameterDef` field:""")
    return


@app.cell
def _(my_i32):
    my_i32.width
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""
        ### Operation Definition

        Operations are defined similarly to `ParametrizedAttribute`, by using the `irdl_op_definition` decorator. The decorator allows the definition of expected operands, results, attributes, and regions. Each definition should contain a `name` static field, which is used for parsing and printing:
        """
    )
    return


@app.cell
def _(printer):
    from xdsl.irdl import IRDLOperation, irdl_op_definition


    @irdl_op_definition
    class MyEmptyOp(IRDLOperation):
        name = "my_dialect.my_op"


    my_op = MyEmptyOp.build()
    printer.print_op(my_op)
    return IRDLOperation, MyEmptyOp, irdl_op_definition, my_op


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""
        #### Operands and Results

        Operands and results are added using fields containing `Operand` and `OpResult`, which each contain an attribute constraint. The order correspond to the operand and result order, and the constraint applies on the SSA variable type.

        Here is an example of an operation defining operands and a result:
        """
    )
    return


@app.cell
def _(IRDLOperation, IntegerAttr, i32, irdl_op_definition, printer):
    from typing import Annotated

    from xdsl.dialects.arith import ConstantOp
    from xdsl.ir import OpResult
    from xdsl.irdl import Operand, operand_def, result_def


    @irdl_op_definition
    class Addi32Op(IRDLOperation):
        name = "arith.addi32"

        # Definition of operands and results.
        # Note the use of the attribute constraint coercion.
        input1 = operand_def(i32)
        input2 = operand_def(i32)
        output = result_def(i32)


    i32_ssa_var = ConstantOp(IntegerAttr.from_int_and_width(62, 32), i32)
    my_addi32 = Addi32Op.build(
        operands=[i32_ssa_var.result, i32_ssa_var.result], result_types=[i32]
    )
    printer.print_op(i32_ssa_var)
    print()
    printer.print_op(my_addi32)
    return (
        Addi32Op,
        Annotated,
        ConstantOp,
        OpResult,
        Operand,
        i32_ssa_var,
        my_addi32,
        operand_def,
        result_def,
    )


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""The operation expects exactly the number of operands and results defined and checks that operands and results satisfy their invariants using the `verify` method.""")
    return


@app.cell
def _(my_addi32):
    # This will pass since the invariants are satisfied
    my_addi32.verify()
    return


@app.cell
def _(Addi32, i32, i32_ssa_var):
    # Wrong number of operands
    try:
        bad_addi32_a = Addi32.build(operands=[i32_ssa_var], result_types=[i32])
    except Exception as e:
        print(e)
    return (bad_addi32_a,)


@app.cell
def _(Addi32Op, i32, i32_ssa_var):
    # Wrong number of results
    try:
        bad_addi32_b = Addi32Op.build(
            operands=[i32_ssa_var, i32_ssa_var], result_types=[i32, i32]
        )
    except Exception as e:
        print(e)
    return (bad_addi32_b,)


@app.cell
def _(Addi32Op, i32_ssa_var, i64):
    # Wrong result type
    bad_addi32_c = Addi32Op.build(operands=[i32_ssa_var, i32_ssa_var], result_types=[i64])

    try:
        bad_addi32_c.verify()
    except Exception as e:
        print(e)
    return (bad_addi32_c,)


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""As with `ParametrizedAttribute`, each operand and result definition defines accessors to easily access operands and results:""")
    return


@app.cell
def _(my_addi32):
    assert my_addi32.input1 == my_addi32.operands[0]
    assert my_addi32.input2 == my_addi32.operands[1]
    assert my_addi32.output == my_addi32.results[0]
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""
        #### Constraint Variables

        Constraint variables can directly be used in Operation and Attribute definitions by using a `TypeAlias` annotated with a `ConstraintVar`.
        """
    )
    return


@app.cell
def _(
    IRDLOperation,
    IntegerType,
    VarConstraint,
    i32,
    i32_ssa_var,
    i64,
    irdl_op_definition,
    operand_def,
    result_def,
):
    from typing import ClassVar
    from xdsl.irdl import base


    @irdl_op_definition
    class BinaryOp(IRDLOperation):
        name = "binary_op"

        T: ClassVar = VarConstraint("T", base(IntegerType))

        lhs = operand_def(T)
        rhs = operand_def(T)
        result = result_def(T)


    op = BinaryOp.build(operands=[i32_ssa_var, i32_ssa_var], result_types=[i32])
    op.verify()

    op_incorrect = BinaryOp.build(operands=[i32_ssa_var, i32_ssa_var], result_types=[i64])
    try:
        op_incorrect.verify()
    except Exception as e:
        print(e)
    return BinaryOp, ClassVar, base, op, op_incorrect


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""
        #### Variadic Operands and Results

        Operand and result definitions can be defined variadic, meaning that their definition can have different numbers of operands or results. Variadic definitions are defined with `VarOperand` and `VarResultDef`.
        """
    )
    return


@app.cell
def _(
    ConstantOp,
    IRDLOperation,
    IntegerAttr,
    i32,
    irdl_op_definition,
    printer,
    result_def,
):
    from xdsl.irdl import VarOperand, var_operand_def


    @irdl_op_definition
    class AddVariadicOp(IRDLOperation):
        name = "add_variadic"
        ops = var_operand_def(i32)
        res = result_def(i32)


    i32_ssa_var_b = ConstantOp(IntegerAttr.from_int_and_width(62, 32), i32)
    add_op = AddVariadicOp.build(operands=[[i32_ssa_var_b] * 3], result_types=[i32])
    printer.print_op(i32_ssa_var_b)
    print()
    printer.print_op(add_op)
    return AddVariadicOp, VarOperand, add_op, i32_ssa_var_b, var_operand_def


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""Generated accessors return a list of SSA variables instead of a single variable:""")
    return


@app.cell
def _(add_op):
    print(len(add_op.ops))
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""Since it may be ambiguous, it is not possible to simply define two variadic operands, or two variadic results. To do so, the operation definition requires the `AttrSizedOperandSegments` or `AttrSizedResultSegments` IRDL option, which require the operation to contain a `operandSegmentSizes` or `resultSegmentSizes` attribute, containing the size of the variadic operands, and variadic results.""")
    return


@app.cell
def _(
    ConstantOp,
    IRDLOperation,
    IntegerAttr,
    i32,
    irdl_op_definition,
    result_def,
    var_operand_def,
):
    from xdsl.dialects.builtin import VectorType
    from xdsl.irdl import AttrSizedOperandSegments


    @irdl_op_definition
    class AddVariadic2Op(IRDLOperation):
        name = "add_variadic"
        ops1 = var_operand_def(i32)
        ops2 = var_operand_def(i32)
        res = result_def(i32)

        irdl_options = [AttrSizedOperandSegments()]


    i32_ssa_var_c = ConstantOp(IntegerAttr.from_int_and_width(62, 32), i32)
    add_op2 = AddVariadic2Op.build(
        operands=[[i32_ssa_var_c] * 2, [i32_ssa_var_c]],
        result_types=[i32],
        attributes={"operandSegmentSizes": VectorType(i32, [2, 1])},
    )
    print("Length of add_op2.ops1:", len(add_op2.ops1))
    print("Length of add_op2.ops2:", len(add_op2.ops2))
    return (
        AddVariadic2Op,
        AttrSizedOperandSegments,
        VectorType,
        add_op2,
        i32_ssa_var_c,
    )


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""In addition to variadic operands and results, IRDL also allows the definition of optional operands and results. Optional operands and results are essentially variadic operands and regions that are constrained to have either 0 or 1 elements. They are defined using `OptOperand` and `OptResultDef`, and define accessors returning optional SSA variables.""")
    return


@app.cell
def _(
    ConstantOp,
    IRDLOperation,
    IntegerAttr,
    i32,
    irdl_op_definition,
    operand_def,
    result_def,
):
    from xdsl.irdl import OptOperand, opt_operand_def


    @irdl_op_definition
    class AddVariadic2Op2(IRDLOperation):
        name = "add_optional"
        ops1 = operand_def(i32)
        ops2 = opt_operand_def(i32)
        res = result_def(i32)


    i32_ssa_var_d = ConstantOp(IntegerAttr.from_int_and_width(62, 32), i32)
    add_op3 = AddVariadic2Op2.build(
        operands=[i32_ssa_var_d, [i32_ssa_var_d]], result_types=[i32]
    )
    print(add_op3.ops2)

    add_op4 = AddVariadic2Op2.build(operands=[i32_ssa_var_d, []], result_types=[i32])
    print(add_op4.ops2)
    return (
        AddVariadic2Op2,
        OptOperand,
        add_op3,
        add_op4,
        i32_ssa_var_d,
        opt_operand_def,
    )


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""
        #### Attributes Definition

        Attribute definitions are defined using `OpAttr`. The field name correspond to the expected attribute name.
        """
    )
    return


@app.cell
def _(IRDLOperation, StringAttr, irdl_op_definition, printer):
    from xdsl.irdl import attr_def


    @irdl_op_definition
    class StringAttrOp(IRDLOperation):
        name = "string_attr_op"
        value = attr_def(StringAttr)


    my_attr_op = StringAttrOp.build(attributes={"value": StringAttr("ga")})
    my_attr_op.verify()
    printer.print_op(my_attr_op)
    return StringAttrOp, attr_def, my_attr_op


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""The defined verifier ensures that the attribute is present:""")
    return


@app.cell
def _(StringAttrOp):
    my_attr_op2 = StringAttrOp.build()
    try:
        my_attr_op2.verify()
    except Exception as e:
        print(e)
    return (my_attr_op2,)


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""And expects that the attribute respect the constraint:""")
    return


@app.cell
def _(IntAttr, StringAttrOp):
    try:
        my_attr_op3 = StringAttrOp.build(attributes={"value": IntAttr(42)})
    except Exception as e:
        print(e)
    return (my_attr_op3,)


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""Note that contrary to operands and results, other attributes may appear in an operation, even though they are not defined by the operation:""")
    return


@app.cell
def _(IntAttr, StringAttr, StringAttrOp, printer):
    my_attr_op4 = StringAttrOp.build(
        attributes={"value": StringAttr("ga"), "other_attr": IntAttr(42)}
    )
    my_attr_op4.verify()
    printer.print_op(my_attr_op4)
    return (my_attr_op4,)


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""
        #### Regions

        Regions definitions are defined using `Region` or `SingleBlockRegion` annotations. The second definition constrains the region to contain a single block, and both definitions allows to further constraint the region by giving a constraint for the entry basic block parameters.
        """
    )
    return


@app.cell
def _(IRDLOperation, i32, irdl_op_definition, printer):
    from xdsl.irdl import Block, Region, region_def, traits_def
    from xdsl.traits import NoTerminator


    @irdl_op_definition
    class WhileOp(IRDLOperation):
        name = "while_op"
        value = region_def()
        traits = traits_def(NoTerminator())


    region = Region(Block(arg_types=[i32]))
    region_op = WhileOp.build(regions=[region])
    region_op.verify()
    printer.print_op(region_op)
    return (
        Block,
        NoTerminator,
        Region,
        WhileOp,
        region,
        region_def,
        region_op,
        traits_def,
    )


@app.cell(hide_code=True)
def _(mo):
    mo.md(
        r"""
        #### Additional Verifiers

        `irdl_op_definition` is not expressive enough to define arbitrary constraints, especially for constraints spanning over multiple operand and result definitions. To circumvent that, definitions may define a `verify_` method that will be called in the generated verifier:
        """
    )
    return


@app.cell
def _(
    ConstantOp,
    IRDLOperation,
    IntegerAttr,
    IntegerType,
    i32,
    i32_ssa_var,
    i64,
    irdl_op_definition,
    operand_def,
    result_def,
):
    from xdsl.dialects.arith import AddiOp


    @irdl_op_definition
    class MyAddiOp(IRDLOperation):
        name = "std.addi"
        input1 = operand_def(IntegerType)
        input2 = operand_def(IntegerType)
        output = result_def(IntegerType)

        # Ensure that the inputs and outputs have the same type:
        def verify_(self) -> None:
            if self.input1.type != self.input2.type or self.input2.type != self.output.type:
                raise Exception("expect all input and output types to be equal")


    i32_ssa_var_e = ConstantOp(IntegerAttr.from_int_and_width(62, 32), i32)
    add_op5 = AddiOp.build(operands=[i32_ssa_var_e, i32_ssa_var_e], result_types=[i32])
    # This will pass, since all operands and results have the same type
    add_op5.verify()

    # This will raise an error, since the result has a different type than the operands
    bad_add_op = AddiOp.build(operands=[i32_ssa_var, i32_ssa_var], result_types=[i64])
    try:
        bad_add_op.verify()
    except Exception as e:
        print(e)
    return AddiOp, MyAddiOp, add_op5, bad_add_op, i32_ssa_var_e


if __name__ == "__main__":
    app.run()



import pytest

from xdsl.utils.disjoint_set import DisjointSet, IntDisjointSet


def test_disjoint_set_init():
    ds = IntDisjointSet(size=5)
    assert ds.value_count() == 5
    # Each element should start in its own set
    for i in range(5):
        assert ds[i] == i


def test_disjoint_set_add():
    ds = IntDisjointSet(size=2)
    assert ds.value_count() == 2

    new_val = ds.add()
    assert new_val == 2
    assert ds.value_count() == 3
    assert ds[new_val] == new_val


def test_disjoint_set_find_invalid():
    ds = IntDisjointSet(size=3)
    with pytest.raises(KeyError):
        ds[3]
    with pytest.raises(KeyError):
        ds[-1]


def test_disjoint_set_union():
    ds = IntDisjointSet(size=4)

    # Union 0 and 1
    assert ds.union(0, 1)
    root = ds[0]
    assert ds[1] == root
    assert ds.connected(0, 1)
    assert not ds.connected(0, 2)

    # Union 2 and 3
    assert ds.union(2, 3)
    root2 = ds[2]
    assert ds[3] == root2
    assert ds.connected(2, 3)
    assert not ds.connected(1, 2)

    # Union already connected elements
    assert not ds.union(0, 1)
    assert ds.connected(0, 1)

    # Union two sets
    assert ds.union(1, 2)
    final_root = ds[0]
    assert ds[1] == final_root
    assert ds[2] == final_root
    assert ds[3] == final_root
    # After unioning all elements, they should all be connected
    assert ds.connected(0, 1)
    assert ds.connected(1, 2)
    assert ds.connected(2, 3)
    assert ds.connected(0, 3)


def test_disjoint_set_path_compression():
    ds = IntDisjointSet(size=4)

    # Create a chain: 3->2->1->0
    ds._parent = [0, 0, 1, 2]  # pyright: ignore[reportPrivateUsage]
    ds._count = [4, 3, 2, 1]  # pyright: ignore[reportPrivateUsage]

    # Find should compress the path
    root = ds[3]
    # After compression, all nodes should point directly to root
    assert ds._parent[3] == root  # pyright: ignore[reportPrivateUsage]
    assert ds._parent[2] == root  # pyright: ignore[reportPrivateUsage]
    assert ds._parent[1] == root  # pyright: ignore[reportPrivateUsage]
    assert ds._parent[0] == root  # pyright: ignore[reportPrivateUsage]


def test_generic_disjoint_set():
    ds = DisjointSet(["a", "b", "c", "d"])

    # Union a and b
    assert ds.union("a", "b")
    root = ds.find("a")
    assert ds.find("b") == root
    assert ds.connected("a", "b")
    assert not ds.connected("a", "c")

    # Union c and d
    assert ds.union("c", "d")
    root2 = ds.find("c")
    assert ds.find("d") == root2
    assert ds.connected("c", "d")
    assert not ds.connected("b", "c")

    # Union already connected elements
    assert not ds.union("a", "b")
    assert ds.connected("a", "b")

    # Union two sets
    assert ds.union("b", "c")
    final_root = ds.find("a")
    assert ds.find("b") == final_root
    assert ds.find("c") == final_root
    assert ds.find("d") == final_root
    # After unioning all elements, they should all be connected
    assert ds.connected("a", "b")
    assert ds.connected("b", "c")
    assert ds.connected("c", "d")
    assert ds.connected("a", "d")


def test_generic_disjoint_set_add():
    ds = DisjointSet(["a", "b"])
    ds.add("c")
    ds.add("d")

    assert ds.union("a", "c")
    root = ds.find("a")
    assert ds.find("c") == root

    assert ds.union("b", "d")
    root2 = ds.find("b")
    assert ds.find("d") == root2


def test_generic_disjoint_set_find_invalid():
    ds = DisjointSet(["a", "b", "c"])
    with pytest.raises(KeyError):
        ds.find("d")



import pytest

from xdsl.dialects.builtin import (
    IndexType,
    IntAttr,
    MemRefType,
    VectorType,
    i1,
    i32,
    i64,
)
from xdsl.dialects.vector import (
    BroadcastOp,
    CreateMaskOp,
    ExtractElementOp,
    FMAOp,
    InsertElementOp,
    LoadOp,
    MaskedLoadOp,
    MaskedStoreOp,
    PrintOp,
    StoreOp,
)
from xdsl.ir import Attribute, OpResult
from xdsl.utils.test_value import TestSSAValue


def get_MemRef_SSAVal(
    referenced_type: Attribute, shape: list[int | IntAttr]
) -> TestSSAValue:
    memref_type = MemRefType(referenced_type, shape)
    return TestSSAValue(memref_type)


def get_Vector_SSAVal(
    referenced_type: Attribute, shape: list[int | IntAttr]
) -> TestSSAValue:
    vector_type = VectorType(referenced_type, shape)
    return TestSSAValue(vector_type)


def test_vectorType():
    vec = VectorType(i32, [1])

    assert vec.get_num_dims() == 1
    assert vec.get_shape() == (1,)
    assert vec.element_type is i32


def test_vectorType_with_dimensions():
    vec = VectorType(i32, [3, 3, 3])

    assert vec.get_num_dims() == 3
    assert vec.get_shape() == (3, 3, 3)
    assert vec.element_type is i32


def test_vector_load_i32():
    memref_ssa_value = get_MemRef_SSAVal(i32, [1])
    load = LoadOp.get(memref_ssa_value, [])

    assert type(load.results[0]) is OpResult
    assert type(load.results[0].type) is VectorType
    assert load.indices == ()


def test_vector_load_i32_with_dimensions():
    memref_ssa_value = get_MemRef_SSAVal(i32, [2, 3])
    index1 = TestSSAValue(IndexType())
    index2 = TestSSAValue(IndexType())
    load = LoadOp.get(memref_ssa_value, [index1, index2])

    assert type(load.results[0]) is OpResult
    assert type(load.results[0].type) is VectorType
    assert load.indices[0] is index1
    assert load.indices[1] is index2


def test_vector_load_verify_type_matching():
    res_vector_type = VectorType(i64, [1])

    memref_ssa_value = get_MemRef_SSAVal(i32, [4, 5])

    load = LoadOp.build(operands=[memref_ssa_value, []], result_types=[res_vector_type])

    with pytest.raises(
        Exception, match="MemRef element type should match the Vector element type."
    ):
        load.verify()


def test_vector_load_verify_indexing_exception():
    memref_ssa_value = get_MemRef_SSAVal(i32, [2, 3])

    load = LoadOp.get(memref_ssa_value, [])

    with pytest.raises(Exception, match="Expected an index for each dimension."):
        load.verify()


def test_vector_store_i32():
    vector_ssa_value = get_Vector_SSAVal(i32, [1])
    memref_ssa_value = get_MemRef_SSAVal(i32, [1])

    store = StoreOp.get(vector_ssa_value, memref_ssa_value, [])

    assert store.base is memref_ssa_value
    assert store.vector is vector_ssa_value
    assert store.indices == ()


def test_vector_store_i32_with_dimensions():
    vector_ssa_value = get_Vector_SSAVal(i32, [2, 3])
    memref_ssa_value = get_MemRef_SSAVal(i32, [4, 5])

    index1 = TestSSAValue(IndexType())
    index2 = TestSSAValue(IndexType())
    store = StoreOp.get(vector_ssa_value, memref_ssa_value, [index1, index2])

    assert store.base is memref_ssa_value
    assert store.vector is vector_ssa_value
    assert store.indices[0] is index1
    assert store.indices[1] is index2


def test_vector_store_verify_type_matching():
    vector_ssa_value = get_Vector_SSAVal(i64, [2, 3])
    memref_ssa_value = get_MemRef_SSAVal(i32, [4, 5])

    store = StoreOp.get(vector_ssa_value, memref_ssa_value, [])

    with pytest.raises(
        Exception, match="MemRef element type should match the Vector element type."
    ):
        store.verify()


def test_vector_store_verify_indexing_exception():
    vector_ssa_value = get_Vector_SSAVal(i32, [2, 3])
    memref_ssa_value = get_MemRef_SSAVal(i32, [4, 5])

    store = StoreOp.get(vector_ssa_value, memref_ssa_value, [])

    with pytest.raises(Exception, match="Expected an index for each dimension."):
        store.verify()


def test_vector_broadcast():
    index1 = TestSSAValue(IndexType())
    broadcast = BroadcastOp.get(index1)

    assert type(broadcast.results[0]) is OpResult
    assert type(broadcast.results[0].type) is VectorType
    assert broadcast.source is index1


def test_vector_broadcast_verify_type_matching():
    index1 = TestSSAValue(IndexType())
    res_vector_type = VectorType(i64, [1])

    broadcast = BroadcastOp.build(operands=[index1], result_types=[res_vector_type])

    with pytest.raises(
        Exception,
        match="Source operand and result vector must have the same element type.",
    ):
        broadcast.verify()


def test_vector_fma():
    i32_vector_type = VectorType(i32, [1])

    lhs_vector_ssa_value = TestSSAValue(i32_vector_type)
    rhs_vector_ssa_value = TestSSAValue(i32_vector_type)
    acc_vector_ssa_value = TestSSAValue(i32_vector_type)

    fma = FMAOp.get(lhs_vector_ssa_value, rhs_vector_ssa_value, acc_vector_ssa_value)

    assert type(fma.results[0]) is OpResult
    assert type(fma.results[0].type) is VectorType
    assert fma.lhs is lhs_vector_ssa_value
    assert fma.rhs is rhs_vector_ssa_value
    assert fma.acc is acc_vector_ssa_value


def test_vector_fma_with_dimensions():
    i32_vector_type = VectorType(i32, [2, 3])

    lhs_vector_ssa_value = TestSSAValue(i32_vector_type)
    rhs_vector_ssa_value = TestSSAValue(i32_vector_type)
    acc_vector_ssa_value = TestSSAValue(i32_vector_type)

    fma = FMAOp.get(lhs_vector_ssa_value, rhs_vector_ssa_value, acc_vector_ssa_value)

    assert type(fma.results[0]) is OpResult
    assert type(fma.results[0].type) is VectorType
    assert fma.lhs is lhs_vector_ssa_value
    assert fma.rhs is rhs_vector_ssa_value
    assert fma.acc is acc_vector_ssa_value


def test_vector_masked_load():
    memref_ssa_value = get_MemRef_SSAVal(i32, [1])
    mask_vector_ssa_value = get_Vector_SSAVal(i1, [1])
    passthrough_vector_ssa_value = get_Vector_SSAVal(i32, [1])

    maskedload = MaskedLoadOp.get(
        memref_ssa_value, [], mask_vector_ssa_value, passthrough_vector_ssa_value
    )

    assert type(maskedload.results[0]) is OpResult
    assert type(maskedload.results[0].type) is VectorType
    assert maskedload.indices == ()


def test_vector_masked_load_with_dimensions():
    memref_ssa_value = get_MemRef_SSAVal(i32, [4, 5])
    mask_vector_ssa_value = get_Vector_SSAVal(i1, [1])
    passthrough_vector_ssa_value = get_Vector_SSAVal(i32, [1])

    index1 = TestSSAValue(IndexType())
    index2 = TestSSAValue(IndexType())

    maskedload = MaskedLoadOp.get(
        memref_ssa_value,
        [index1, index2],
        mask_vector_ssa_value,
        passthrough_vector_ssa_value,
    )

    assert type(maskedload.results[0]) is OpResult
    assert type(maskedload.results[0].type) is VectorType
    assert maskedload.indices[0] is index1
    assert maskedload.indices[1] is index2


def test_vector_masked_load_verify_memref_res_type_matching():
    memref_ssa_value = get_MemRef_SSAVal(i32, [1])
    mask_vector_ssa_value = get_Vector_SSAVal(i1, [1])
    passthrough_vector_ssa_value = get_Vector_SSAVal(i32, [1])

    i64_res_vector_type = VectorType(i64, [1])

    maskedload = MaskedLoadOp.build(
        operands=[
            memref_ssa_value,
            [],
            mask_vector_ssa_value,
            passthrough_vector_ssa_value,
        ],
        result_types=[i64_res_vector_type],
    )

    message = (
        "MemRef element type should match the result vector and passthrough "
        "vector element type. Found different element types for memref and result."
    )
    with pytest.raises(Exception, match=message):
        maskedload.verify()


def test_vector_masked_load_verify_memref_passthrough_type_matching():
    memref_ssa_value = get_MemRef_SSAVal(i32, [1])
    mask_vector_ssa_value = get_Vector_SSAVal(i1, [1])
    passthrough_vector_ssa_value = get_Vector_SSAVal(i64, [1])

    i64_res_vector_type = VectorType(i32, [1])

    maskedload = MaskedLoadOp.build(
        operands=[
            memref_ssa_value,
            [],
            mask_vector_ssa_value,
            passthrough_vector_ssa_value,
        ],
        result_types=[i64_res_vector_type],
    )

    message = (
        "MemRef element type should match the result vector and passthrough "
        "vector element type. Found different element types for memref and passthrough."
    )

    with pytest.raises(Exception, match=message):
        maskedload.verify()


def test_vector_masked_load_verify_indexing_exception():
    memref_ssa_value = get_MemRef_SSAVal(i32, [4, 5])
    mask_vector_ssa_value = get_Vector_SSAVal(i1, [2])
    passthrough_vector_ssa_value = get_Vector_SSAVal(i32, [1])

    maskedload = MaskedLoadOp.get(
        memref_ssa_value, [], mask_vector_ssa_value, passthrough_vector_ssa_value
    )

    with pytest.raises(Exception, match="Expected an index for each memref dimension."):
        maskedload.verify()


def test_vector_masked_store():
    memref_ssa_value = get_MemRef_SSAVal(i32, [1])
    mask_vector_ssa_value = get_Vector_SSAVal(i1, [1])
    value_to_store_vector_ssa_value = get_Vector_SSAVal(i32, [1])

    maskedstore = MaskedStoreOp.get(
        memref_ssa_value, [], mask_vector_ssa_value, value_to_store_vector_ssa_value
    )

    assert maskedstore.base is memref_ssa_value
    assert maskedstore.mask is mask_vector_ssa_value
    assert maskedstore.value_to_store is value_to_store_vector_ssa_value
    assert maskedstore.indices == ()


def test_vector_masked_store_with_dimensions():
    memref_ssa_value = get_MemRef_SSAVal(i32, [4, 5])
    mask_vector_ssa_value = get_Vector_SSAVal(i1, [1])
    value_to_store_vector_ssa_value = get_Vector_SSAVal(i32, [1])

    index1 = TestSSAValue(IndexType())
    index2 = TestSSAValue(IndexType())

    maskedstore = MaskedStoreOp.get(
        memref_ssa_value,
        [index1, index2],
        mask_vector_ssa_value,
        value_to_store_vector_ssa_value,
    )

    assert maskedstore.base is memref_ssa_value
    assert maskedstore.mask is mask_vector_ssa_value
    assert maskedstore.value_to_store is value_to_store_vector_ssa_value
    assert maskedstore.indices[0] is index1
    assert maskedstore.indices[1] is index2


def test_vector_masked_store_verify_memref_value_to_store_type_matching():
    memref_ssa_value = get_MemRef_SSAVal(i32, [1])
    mask_vector_ssa_value = get_Vector_SSAVal(i1, [1])
    value_to_store_vector_ssa_value = get_Vector_SSAVal(i64, [1])

    maskedstore = MaskedStoreOp.get(
        memref_ssa_value, [], mask_vector_ssa_value, value_to_store_vector_ssa_value
    )

    message = (
        "MemRef element type should match the stored vector type. "
        "Obtained types were i32 and i64."
    )
    with pytest.raises(Exception, match=message):
        maskedstore.verify()


def test_vector_masked_store_verify_indexing_exception():
    memref_ssa_value = get_MemRef_SSAVal(i32, [4, 5])
    mask_vector_ssa_value = get_Vector_SSAVal(i1, [2])
    value_to_store_vector_ssa_value = get_Vector_SSAVal(i32, [1])

    maskedstore = MaskedStoreOp.get(
        memref_ssa_value, [], mask_vector_ssa_value, value_to_store_vector_ssa_value
    )

    with pytest.raises(Exception, match="Expected an index for each memref dimension."):
        maskedstore.verify()


def test_vector_print():
    vector_ssa_value = get_Vector_SSAVal(i32, [1])

    print = PrintOp.get(vector_ssa_value)

    assert print.source is vector_ssa_value


def test_vector_create_mask():
    create_mask = CreateMaskOp.get([])

    assert type(create_mask.results[0]) is OpResult
    assert type(create_mask.results[0].type) is VectorType
    assert create_mask.mask_dim_sizes == ()


def test_vector_create_mask_with_dimensions():
    index1 = TestSSAValue(IndexType())
    index2 = TestSSAValue(IndexType())

    create_mask = CreateMaskOp.get([index1, index2])

    assert type(create_mask.results[0]) is OpResult
    assert type(create_mask.results[0].type) is VectorType
    assert create_mask.mask_dim_sizes[0] is index1
    assert create_mask.mask_dim_sizes[1] is index2


def test_vector_create_mask_verify_indexing_exception():
    mask_vector_type = VectorType(i1, [2, 3])

    create_mask = CreateMaskOp.build(operands=[[]], result_types=[mask_vector_type])

    with pytest.raises(
        Exception,
        match="Expected an operand value for each dimension of resultant mask.",
    ):
        create_mask.verify()


def test_vector_extract_element_verify_vector_rank_0_or_1():
    vector_type = VectorType(IndexType(), [3, 3])

    vector = TestSSAValue(vector_type)
    position = TestSSAValue(IndexType())
    extract_element = ExtractElementOp(vector, position)

    with pytest.raises(Exception, match="Unexpected >1 vector rank."):
        extract_element.verify()


def test_vector_extract_element_construction_1d():
    vector_type = VectorType(IndexType(), [3])

    vector = TestSSAValue(vector_type)
    position = TestSSAValue(IndexType())

    extract_element = ExtractElementOp(vector, position)

    assert extract_element.vector is vector
    assert extract_element.position is position
    assert extract_element.result.type == vector_type.element_type


def test_vector_extract_element_1d_verify_non_empty_position():
    vector_type = VectorType(IndexType(), [3])

    vector = TestSSAValue(vector_type)

    extract_element = ExtractElementOp(vector)

    with pytest.raises(Exception, match="Expected position for 1-D vector."):
        extract_element.verify()


def test_vector_extract_element_construction_0d():
    vector_type = VectorType(IndexType(), [])

    vector = TestSSAValue(vector_type)

    extract_element = ExtractElementOp(vector)

    assert extract_element.vector is vector
    assert extract_element.position is None
    assert extract_element.result.type == vector_type.element_type


def test_vector_extract_element_0d_verify_empty_position():
    vector_type = VectorType(IndexType(), [])

    vector = TestSSAValue(vector_type)
    position = TestSSAValue(IndexType())

    extract_element = ExtractElementOp(vector, position)

    with pytest.raises(
        Exception, match="Expected position to be empty with 0-D vector."
    ):
        extract_element.verify()


def test_vector_insert_element_verify_vector_rank_0_or_1():
    vector_type = VectorType(IndexType(), [3, 3])

    source = TestSSAValue(IndexType())
    dest = TestSSAValue(vector_type)
    position = TestSSAValue(IndexType())

    insert_element = InsertElementOp(source, dest, position)

    with pytest.raises(Exception, match="Unexpected >1 vector rank."):
        insert_element.verify()


def test_vector_insert_element_construction_1d():
    vector_type = VectorType(IndexType(), [3])

    source = TestSSAValue(IndexType())
    dest = TestSSAValue(vector_type)
    position = TestSSAValue(IndexType())

    insert_element = InsertElementOp(source, dest, position)

    assert insert_element.source is source
    assert insert_element.dest is dest
    assert insert_element.position is position
    assert insert_element.result.type == vector_type


def test_vector_insert_element_1d_verify_non_empty_position():
    vector_type = VectorType(IndexType(), [3])

    source = TestSSAValue(IndexType())
    dest = TestSSAValue(vector_type)

    insert_element = InsertElementOp(source, dest)

    with pytest.raises(
        Exception,
        match="Expected position for 1-D vector.",
    ):
        insert_element.verify()


def test_vector_insert_element_construction_0d():
    vector_type = VectorType(IndexType(), [])

    source = TestSSAValue(IndexType())
    dest = TestSSAValue(vector_type)

    insert_element = InsertElementOp(source, dest)

    assert insert_element.source is source
    assert insert_element.dest is dest
    assert insert_element.position is None
    assert insert_element.result.type == vector_type


def test_vector_insert_element_0d_verify_empty_position():
    vector_type = VectorType(IndexType(), [])

    source = TestSSAValue(IndexType())
    dest = TestSSAValue(vector_type)
    position = TestSSAValue(IndexType())

    insert_element = InsertElementOp(source, dest, position)

    with pytest.raises(
        Exception,
        match="Expected position to be empty with 0-D vector.",
    ):
        insert_element.verify()



from abc import ABC, abstractmethod
from collections.abc import Sequence
from contextlib import contextmanager
from dataclasses import dataclass
from typing import Generic, TypeVar

from xdsl.backend.register_type import RegisterType

_T = TypeVar("_T", bound=RegisterType)


@dataclass
class RegisterQueue(Generic[_T], ABC):
    """
    LIFO queue of registers available for allocation.
    """

    @abstractmethod
    def push(self, reg: _T) -> None:
        """
        Return a register to be made available for allocation.
        """
        ...

    @abstractmethod
    def pop(self, reg_type: type[_T]) -> _T:
        """
        Get the next available register for allocation.
        """
        ...

    @contextmanager
    def reserve_registers(self, regs: Sequence[_T]):
        for reg in regs:
            self.reserve_register(reg)

        yield

        for reg in regs:
            self.unreserve_register(reg)

    @abstractmethod
    def reserve_register(self, reg: _T) -> None:
        """
        Increase the reservation count for a register.
        If the reservation count is greater than 0, a register cannot be pushed back onto
        the queue.
        It is invalid to reserve a register that is available, and popping it before
        unreserving a register will result in an AssertionError.
        """
        ...

    @abstractmethod
    def unreserve_register(self, reg: _T) -> None:
        """
        Decrease the reservation count for a register. If the reservation count is 0, make
        the register available for allocation.
        """
        ...

    @abstractmethod
    def exclude_register(self, reg: _T) -> None:
        """
        Removes regiesters from available set, if present.
        """
        ...
